# -*- coding: utf-8 -*-
xhtmlInput='''
<table id="configabsences_cal" class="list timeline" width="90%">
      <colgroup>
       <col width="100px"/>
       <col style=""/><col style=""/><col style=""/><col style=""/><col style="background-color: #dedede"/><col style="background-color: #c0c0c0"/><col style="background-color: #c0c0c0"/><col style=""/><col style=""/><col style=""/><col style=""/><col style="background-color: #dedede"/><col style="background-color: #c0c0c0"/><col style="background-color: #c0c0c0"/><col style=""/><col style=""/><col style=""/><col style=""/><col style="background-color: #dedede"/><col style="background-color: #c0c0c0"/><col style="background-color: #c0c0c0"/><col style=""/><col style=""/><col style=""/><col style=""/><col style="background-color: #dedede"/><col style="background-color: #c0c0c0"/><col style="background-color: #c0c0c0"/><col style=""/><col style=""/><col style=""/><col style=""/><col style="background-color: #dedede"/><col style="background-color: #c0c0c0"/><col style="background-color: #c0c0c0"/><col style=""/><col style=""/><col style=""/><col style=""/><col style="background-color: #dedede"/><col style="background-color: #c0c0c0"/><col style="background-color: #c0c0c0"/>
       <col style="width: 75px"/>
      </colgroup>
      <tbody>
     <tr><th></th>
      <th colspan="6">Février 2015</th><th colspan="31">Mars 2015</th><th colspan="5"><acronym title="Avril 2015">A</acronym></th><th></th></tr>
     <tr><td></td> <td><b>L</b></td><td><b>M</b></td><td><b>M</b></td><td><b>J</b></td><td><b>V</b></td><td><b>S</b></td><td><b>D</b></td><td><b>L</b></td><td><b>M</b></td><td><b>M</b></td><td><b>J</b></td><td><b>V</b></td><td><b>S</b></td><td><b>D</b></td><td><b>L</b></td><td><b>M</b></td><td><b>M</b></td><td><b>J</b></td><td><b>V</b></td><td><b>S</b></td><td><b>D</b></td><td><b>L</b></td><td><b>M</b></td><td><b>M</b></td><td><b>J</b></td><td><b>V</b></td><td><b>S</b></td><td><b>D</b></td><td><b>L</b></td><td><b>M</b></td><td><b>M</b></td><td><b>J</b></td><td><b>V</b></td><td><b>S</b></td><td><b>D</b></td><td><b>L</b></td><td><b>M</b></td><td><b>M</b></td><td><b>J</b></td><td><b>V</b></td><td><b>S</b></td><td><b>D</b></td>
      <td></td></tr>
      <tr><td></td> <td><b>23</b></td><td><b>24</b></td><td><b>25</b></td><td><b>26</b></td><td><b>27</b></td><td><b>28</b></td><td><b>01</b></td><td><b>02</b></td><td><b>03</b></td><td><b>04</b></td><td><b>05</b></td><td><b>06</b></td><td><b>07</b></td><td><b>08</b></td><td><b>09</b></td><td><b>10</b></td><td><b>11</b></td><td><b>12</b></td><td><b>13</b></td><td><b>14</b></td><td><b>15</b></td><td><b>16</b></td><td><b>17</b></td><td><b>18</b></td><td><b>19</b></td><td><b>20</b></td><td><b>21</b></td><td><b>22</b></td><td><b>23</b></td><td><b>24</b></td><td><b>25</b></td><td><b>26</b></td><td><b>27</b></td><td><b>28</b></td><td><b>29</b></td><td><b>30</b></td><td><b>31</b></td><td><b>01</b></td><td><b>02</b></td><td><b>03</b></td><td><b>04</b></td><td><b>05</b></td>
       <td></td></tr>
        <tr>
         <td class="tlLeft"><a target="_blank" href="http://localhost:8080/config/GardesUser1350721915326376433960017169/view?page=preferences"><acronym title="Barbason Alain">Alain Barbason</acronym></a></td>
           <td style="background-color: #E5B620">z</td>
           <td style="background-color: #E5B620"></td>
           <td style="background-color: #E5B620"></td>
           <td></td>
           <td style="background-color: #E5B620"></td>
           <td style="background-color: #E5B620"></td>
           <td></td>
           <td></td>
           <td></td>
           <td></td>
           <td></td>
           <td></td>
           <td style="background-color: #13751F"></td>
           <td style="background-color: #13751F"></td>
           <td></td>
           <td></td>
           <td title="Congé (AM), Congrès (PM)" style="background-image: url(http://localhost:8080/ui/angled.png)"></td>
           <td></td>
           <td></td>
           <td style="background-color: #d08181"></td>
           <td style="background-color: #d08181"></td>
           <td style="background-color: #d08181"></td>
           <td style="background-color: #d08181"></td>
           <td style="background-color: #d08181"></td>
           <td style="background-color: #d08181"></td>
           <td style="background-color: #d08181"></td>
           <td style="background-color: #d08181"></td>
           <td style="background-color: #d08181"></td>
           <td></td>
           <td></td>
           <td></td>
           <td></td>
           <td></td>
           <td></td>
           <td></td>
           <td></td>
           <td></td>
           <td></td>
           <td></td>
           <td></td>
           <td></td>
           <td></td>
          <td class="tlRight"><a target="_blank" href="http://localhost:8080/config/GardesUser1350721915326376433960017169/view?page=preferences"><acronym title="Barbason Alain">AB</acronym></a></td>
        </tr><tr>
          <td class="tlLeft"><a target="_blank" href="http://localhost:8080/config/GardesUser1350721915119231834005028903/view?page=preferences"><acronym title="Blom-Peters Lucien">LB</acronym></a></td>
           <td></td>
           <td></td>
           <td></td>
           <td></td>
           <td></td>
           <td></td>
           <td></td>
           <td></td>
           <td></td>
           <td></td>
           <td></td>
           <td></td>
           <td></td>
           <td></td>
           <td></td>
           <td></td>
           <td></td>
           <td></td>
           <td></td>
           <td></td>
           <td></td>
           <td></td>
           <td></td>
           <td></td>
           <td></td>
           <td></td>
           <td style="background-color: #d08181"></td>
           <td style="background-color: #d08181"></td>
           <td style="background-color: #d08181"></td>
           <td style="background-color: #d08181"></td>
           <td style="background-color: #d08181"></td>
           <td style="background-color: #d08181"></td>
           <td style="background-color: #d08181"></td>
           <td style="background-color: #d08181"></td>
           <td style="background-color: #d08181"></td>
           <td></td>
           <td></td>
           <td></td>
           <td></td>
           <td></td>
           <td></td>
           <td></td>
           <td class="tlRight"><a target="_blank" href="http://localhost:8080/config/GardesUser1350721915119231834005028903/view?page=preferences"><acronym title="Blom-Peters Lucien">LB</acronym></a></td>
        </tr>
      </tbody>
     <tbody id="configabsences_trs">
      <script>new AjaxData('configabsences_trs', 'absences:pxTotalRowsFromAjax', {}, 'configabsences')</script>
      <tr>
       <td class="tlLeft">
        <acronym title="Nombre de travailleurs disponibles"><b>P</b></acronym></td>
       <td>42</td><td>42</td><td>41</td><td>42</td><td>41</td><td>39</td><td>40</td><td>42</td><td>41</td><td>41</td><td>41</td><td>42</td><td>37</td><td>37</td><td>41</td><td>42</td><td>40</td><td>39</td><td>39</td><td>37</td><td>37</td><td>39</td><td>37</td><td>36</td><td>36</td><td>36</td><td>31</td><td>32</td><td>38</td><td>39</td><td>39</td><td>39</td><td>38</td><td>37</td><td>37</td><td>42</td><td>41</td><td>41</td><td>41</td><td>42</td><td>33</td><td>33</td>
       <td class="tlRight">
        <acronym title="Nombre de travailleurs disponibles"><b>P</b></acronym></td>
      </tr><tr>
       <td class="tlLeft">
        <acronym title="Nombre total de travailleurs"><b>T</b></acronym></td>
       <td>46</td><td>46</td><td>46</td><td>46</td><td>46</td><td>46</td><td>46</td><td>46</td><td>46</td><td>46</td><td>46</td><td>46</td><td>46</td><td>46</td><td>46</td><td>46</td><td>46</td><td>46</td><td>46</td><td>46</td><td>46</td><td>46</td><td>46</td><td>46</td><td>46</td><td>46</td><td>46</td><td>46</td><td>46</td><td>46</td><td>46</td><td>46</td><td>46</td><td>46</td><td>46</td><td>46</td><td>46</td><td>46</td><td>46</td><td>46</td><td>46</td><td>46</td>
       <td class="tlRight">
        <acronym title="Nombre total de travailleurs"><b>T</b></acronym></td>
      </tr>
     </tbody>
      <tbody> 
       
      <tr><td></td> 
       <td><b>23</b></td><td><b>24</b></td><td><b>25</b></td><td><b>26</b></td><td><b>27</b></td><td><b>28</b></td><td><b>01</b></td><td><b>02</b></td><td><b>03</b></td><td><b>04</b></td><td><b>05</b></td><td><b>06</b></td><td><b>07</b></td><td><b>08</b></td><td><b>09</b></td><td><b>10</b></td><td><b>11</b></td><td><b>12</b></td><td><b>13</b></td><td><b>14</b></td><td><b>15</b></td><td><b>16</b></td><td><b>17</b></td><td><b>18</b></td><td><b>19</b></td><td><b>20</b></td><td><b>21</b></td><td><b>22</b></td><td><b>23</b></td><td><b>24</b></td><td><b>25</b></td><td><b>26</b></td><td><b>27</b></td><td><b>28</b></td><td><b>29</b></td><td><b>30</b></td><td><b>31</b></td><td><b>01</b></td><td><b>02</b></td><td><b>03</b></td><td><b>04</b></td><td><b>05</b></td>
       <td></td></tr>
     <tr><td></td> 
      <td><b>L</b></td><td><b>M</b></td><td><b>M</b></td><td><b>J</b></td><td><b>V</b></td><td><b>S</b></td><td><b>D</b></td><td><b>L</b></td><td><b>M</b></td><td><b>M</b></td><td><b>J</b></td><td><b>V</b></td><td><b>S</b></td><td><b>D</b></td><td><b>L</b></td><td><b>M</b></td><td><b>M</b></td><td><b>J</b></td><td><b>V</b></td><td><b>S</b></td><td><b>D</b></td><td><b>L</b></td><td><b>M</b></td><td><b>M</b></td><td><b>J</b></td><td><b>V</b></td><td><b>S</b></td><td><b>D</b></td><td><b>L</b></td><td><b>M</b></td><td><b>M</b></td><td><b>J</b></td><td><b>V</b></td><td><b>S</b></td><td><b>D</b></td><td><b>L</b></td><td><b>M</b></td><td><b>M</b></td><td><b>J</b></td><td><b>V</b></td><td><b>S</b></td><td><b>D</b></td>
      <td></td></tr>
       
     <tr><th></th> 
      <th colspan="6">Février 2015</th><th colspan="31">Mars 2015</th><th colspan="5"><acronym title="Avril 2015">A</acronym></th><th></th></tr>
      </tbody>
     </table>

<p></p>
<table cellspacing="1" cellpadding="1" border="1" style="width:500px">
  <tbody>
    <tr><td style="width:20%">1</td>
            <td>Maecenas vitae ligula eu nunc ultrices consectetur a ac mi. Cras cursus risus elit. In leo tellus, vehicula nec velit id, rhoncus iaculis leo. Nullam nec dapibus urna. Morbi vulputate dictum turpis. ipsum non finibus.</td>
            <td>Maecenas vitae ligula eu nunc ultrices consectetur a ac mi. Cras cursus risus elit. In leo tellus, vehicula nec velit id, rhoncus iaculis leo. Nullam nec dapibus urna. Morbi vulputate dictum turpis. Vivamus eget consequat nisl. Nam tempus auctor ipsum at lobortis. Vesti laoreet. Cras tristique vitae ipsum non finibus.</td>
        </tr>
    <tr><td>2</td>
            <td>Maecenas vitae ligula eu nunc ultrices consectetur a ac mi. Cras cursus risus elit. In leo tellus, vehicula nec velit id, rhoncus iaculis leo. Nullam nec dapibus urna. Morbi vulputate dictum turpis. Vivamus eget consequat nisl. Nam tempus auctor ipsum at lobortis. Vestibulum ante ipsum primis in faucibus orci luctus et ultrices posuere cubilia Curae; Nam semper ex vitae augue viverra, ut euismod quam laoreet. Cras tristique vitae ipsum non finibus.</td>
            <td>Maecenas vitae ligula eu nunc ultrices consectetur a ac mi. Cras curvitae ipsum non finibus.</td>
        </tr>
    <tr><td>3</td>
            <td>Maecenas vitae ligula eu nunc ultrices consectetur a ac mi. Cras cursus risus elit. In leo tellus, vehicula nec velit id, rhoncus iaculis leo. Nullam nec dapibus urna. Morbi vulputate dictum turpis. Vivamus eget consequat nisl. Nam tempus auctor ipsum at lobortis. Vestibulum ante ipsum primis in faucibus orci luctus et ultrices&nbsp;</td>
            <td>&nbsp;</td>
        </tr>
  </tbody>
</table>
'''



'''A guard is responsible for authenticating and authorizing app users, be they
   humans or other apps. A guard will ensure that any action is performed in the
   respect of security rules defined: workflows, permissions, etc.'''

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
from pathlib import Path

from DateTime import DateTime

from appy.px import Px
from appy.utils import multicall
from appy.model.user import User
from appy.server.cookie import Cookie
from appy.utils import path as putils
from appy.model.utils import Object as O
from appy.model.fields.string import String
from appy.model.fields.password import Password

# Errors - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
KO_LOG_TFM  = 'Invalid value "%s" for attribute "loginTransform".'
KO_PERM     = '%s "%s": "%s" disallowed.'
AUTH_KO     = 'Authentication failed with login %s.'
AUTH_OK     = 'Logged in.'

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Config:
    '''Security configuration'''
    # Allowed value for field "loginTransform"
    transformValues = ('lower', 'upper', 'capitalize')

    def __init__(self):
        # Activate or not the button on home page for asking a new password
        self.activateForgotPassword = True
        # By default, when creating a local user, Appy generates a password for
        # him. If you want to disable this and manage yourself, in your app's
        # User.onEdit method, generation and sending of passwords, set the
        # following attribute to False.
        self.generatePassword = True
        # On user log on, you may want to ask them to choose their
        # "authentication context" in a list. Such a context may have any
        # semantics, is coded as a string and will be accessible on the Handler
        # object in attribute "authContext". In order to activate authentication
        # contexts, place here an instance of a sub-class of class
        # appy.server.context.AuthenticationContext. Consult this class'
        # docstring for more information.
        self.authContext = None
        # Allow self-registering ?
        self.selfRegister = False
        # When a login is encoded by the user, a transform can be applied (one
        # among Config.transformValues).
        self.loginTransform = None
        # When using a LDAP for authenticating users, place an instance of class
        # appy.server.ldap.Config in the field below.
        self.ldap = None
        # When using, in front of an Appy server, a reverse proxy for
        # authentication for achieving single-sign-on (SSO), place an instance
        # of appy.server.sso.Config in the field below.
        self.sso = None
        # If the following attribute is not None, a warning about accepting
        # cookies will be shown until the user accepts it. His acceptance will
        # be stored in a cookie named "cookiesAccepted". Attribute
        # "cookieWarning" must contain the name of an attribute on the tool,
        # that must store the URL of a page displaying the privacy policy.
        self.cookieWarning = False

    def getLdap(self):
        '''Returns the LDAP config if it exists'''
        # The LDAP config can be found either in attribute "ldap" or in the
        # context of a SSO system. This method returns a tuple
        #                     (ldapConfig, context)
        # "context" being None if the global LDAP config is returned, or being
        # the SSO config if the LDAP is the one embedded in the SSO config.
        if self.ldap:
            r = self.ldap, None
        else:
            sso = self.sso
            if not sso or not sso.ldap:
                r = None, None
            else:
                r = sso.ldap, sso
        return r

    def check(self):
        '''Check this config'''
        # Check attribute "loginTransform"
        transform = self.loginTransform
        if transform and (transform not in self.transformValues):
            raise Exception(KO_LOG_TFM % transform)
        return True

    def transformLogin(self, login):
        '''Return p_login as potentially transformed via
           p_self.loginTransform.'''
        return getattr(login, self.loginTransform)() \
               if self.loginTransform else login

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Unauthorized(Exception):
    '''Error that is raised when a security problem is encountered'''

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Guard:
    '''Implements security-related functionality'''
    Error = Unauthorized
    Cookie = Cookie
    traverse = {}

    def __init__(self, handler):
        # The handler that has instantiated this guard
        self.handler = handler
        # Info about the currently selected authentication context
        self.authContext = None
        # A human-readable name for this context
        self.authContextName = None
        # If the authentication context corresponds to a Appy object, the
        # following attribute will get this object once per request and store it
        # here.
        self.authObject = None
        # Unwrap some useful objects
        self.model = handler.server.model
        self.config = handler.server.config
        # Authenticate the currently logged user and get its User instance
        self.user = User.authenticate(self)
        # Remember that this user hits the server now
        if not handler.fake:
            handler.onlineUsers[self.user.login] = DateTime()
        # Cache info about this user
        self.cache()
        # Has the user accepted the cookie warning ?
        if self.config.security.cookieWarning:
            self.cookiesAccepted = handler.req.cookiesAccepted == '1'
        else:
            # We don't care about accepting cookies
            self.cookiesAccepted = True

    def cache(self):
        '''Pre-compute, for performance, heavily requested information about the
           currently logged user.'''
        u = self.user
        self.userLogin = u.login
        self.userLogins = u.getLogins(compute=True, guard=self)
        self.userRoles = u.getRoles(compute=True, guard=self)
        self.userAllowed = u.getAllowedFrom(self.userRoles, self.userLogins)
        self.userLanguage = u.getLanguage() if not self.handler.fake else 'en'

    # In the remaining of this class, when talking about "the user", we mean
    # "the currently logged user".

    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # Security checks on classes
    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    def mayInstantiate(self, class_, checkInitiator=False, initiator=None,
                       raiseOnError=False):
        '''May the user create instances of p_class_ ?'''
        # This information can be defined on p_class_.python, in static
        # attribute "creators". If this attribute holds...
        #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        # a list    | we consider it to be a list of roles, and we check that
        #           | the user has at least one of those roles;
        # a boolean | we consider that the user can create instances of this
        #           | class if the boolean is True;
        # a method  | we execute the method, and via its result, we fall again
        #           | in one of the above cases.
        #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        # If p_class_.python does not define such an attribute, a default list
        # of roles defined in the config will be used.
        #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        # An initiator object may dictate his own rules for instance creation
        if checkInitiator:
            # No need to check for an initiator in the context of a root class
            init = initiator or self.user.getInitiator()
            if init and getattr(init.field, 'creators', None):
                return self.user.hasRole(init.field.creators, init.o)
        # Get the Python class defined in the app
        class_ = class_.python
        # Get attribute "creators" on the class, or use the config
        creators = class_.creators if hasattr(class_, 'creators') \
                   else self.config.model.defaultCreators
        # If "creators" is a method, execute it
        if callable(creators): creators = creators(self.handler.tool)
        # Manage the "boolean" case
        if isinstance(creators, bool) or not creators: return creators
        # Manage the "roles" case
        for role in self.userRoles:
            if role in creators:
                return True

    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # Security checks on objects
    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    def allows(self, o, permission='read', raiseError=False):
        '''Has the user p_permission on p_o ?'''
        r = self.user.hasPermission(permission, o)
        if not r and raiseError:
            raise Unauthorized(KO_PERM % (o.class_.name, o.id, permission))
        return r

    def mayAct(self, o):
        '''m_mayAct allows to hide the whole set of actions for an p_o(bject).
           Indeed, beyond workflow security, it can be useful to hide controls
           like "edit" icons/buttons. For example, if a user may only edit some
           Ref fields with add=True on an object, when clicking on "edit", he
           will see an empty edit form.'''
        return multicall(o, 'mayAct', True)

    def mayDelete(self, o):
        '''May the currently logged user delete this p_o(bject) ?'''
        r = self.allows(o, 'delete')
        if not r: return
        # An additional, user-defined condition, may refine the base permission
        return multicall(o, 'mayDelete', True)

    def mayEdit(self, o, permission='write', permOnly='specific',
                raiseError=False):
        '''May the user edit this object ? p_permission can be a field-specific
           permission. If p_permOnly is True, the additional user-defined
           condition (custom m_mayEdit) is not evaluated. If p_permOnly is
           "specific", this condition is evaluated only if the permission is
           specific (=not "write") If p_raiseError is True, if the user may not
           edit p_o, an error is raised.'''
        r = self.allows(o, permission, raiseError=raiseError)
        if not r: return
        if (permOnly == True) or \
           ((permOnly == 'specific') and (permission != 'write')): return r
        # An additional, user-defined condition, may refine the base permission
        r = multicall(o, 'mayEdit', True)
        if not r and raiseError:
            method = '%s::mayEdit' % o.class_.name
            raise Unauthorized(KO_PERM % (method, o.id, permission))
        return r

    def mayView(self, o, permission='read', raiseError=False):
        '''May the user view this p_o(bject) ? p_permission can be a field-
           specific permission. If p_raiseError is True, if the user may not
           view p_o, an error is raised.'''
        r = self.allows(o, permission, raiseError=raiseError)
        if not r: return
        # An additional, user-defined condition, may refine the base permission
        r = multicall(o, 'mayView', True)
        if not r and raiseError:
            method = '%s::mayView' % o.class_.name
            raise Unauthorized(KO_PERM % (method, o.id, permission))
        return r

    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # Login / logout
    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    traverse['enter'] = True
    def enter(self, tool):
        '''Perform login'''
        user = self.user
        req = tool.req
        handler = tool.H()
        if user.isAnon():
            # Authentication failed
            label = 'login_ko'
            success = False
            login = req.login.strip() if req.login else ''
            msg = AUTH_KO % login
        else:
            # Authentication succeeded
            label = 'login_ok'
            success = True
            msg = AUTH_OK
        # Log the action
        tool.log(msg)
        # Redirect the user to the appropriate page
        if success:
            if user.changePasswordAtNextLogin:
                # Redirect to the page where the user can change its password
                back = '%s/edit?page=password' % user.url
                label = 'login_ok_change_password'
            elif req.goto:
                # The user came to its home page for authenticating: he must now
                # be redirected to the page at request key "goto".
                back = req.goto
            else:
                # Redirect to the app's home page
                back = user.tool.computeHomePage()
                info = req.authInfo
                # Carry custom authentication data when appropriate
                if info: back = '%s?authInfo=%s' % (back, info)
        else:
            # Stay on the same page
            back = tool.H().headers['Referer']
        # Redirect the user to the appropriate page
        tool.resp.goto(back, message=user.translate(label))

    traverse['leave'] = True
    def leave(self, tool):
        '''Perform logout'''
        handler = tool.H()
        # Disable the authentication cookie
        Cookie.disable(handler)
        # Log the action
        handler.log('app', 'info', 'Logged out.')
        # Redirect the user to the app's home page
        back = '%s/public' % tool.url if self.config.ui.discreetLogin \
                                      else self.config.server.getUrl(handler)
        handler.resp.goto(back, message=tool.translate('logout_ok'))

    def getLogoutUrl(self, tool, user):
        '''The "logout" URL may not be the standard Appy one if the app is
           behind a SSO reverse proxy.'''
        if user.source == 'sso': return self.config.security.sso.logoutUrl
        return '%s/guard/leave' % tool.url

    def getLoginUrl(self, ctx):
        '''When discreet logins are in use, compute the JS code or link allowing
           to open the login box as a popup or redirect the user to tool/home,
           depending on the config.'''
        if ctx.cfg.discreetLogin == 'home':
            # "discreet" may hold "home" or "homes". Attribue "stay" will force
            # to stay on this page. Indeed, without it, in "discreet login"
            # mode, tool/home[s] redirects to tool/public.
            r = '%s/%s?stay=1' % (ctx.tool.url, ctx.cfg.home)
        else:
            r = 'javascript:toggleLoginBox(true)'
        return r

    def getSelfRegisterJS(self, tool):
        '''Gets the JS code allowing to create a User instance for the purpose
           of self-registering.'''
        # Compute the URL to post the form to
        url = '%s/new' % tool.url
        return "javascript:post('%s', {'action': 'Create', 'className': " \
               "'User', 'nav': 'ref.tool.users.0.0'})" % url

    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # Password reinitialisation
    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    traverse['askPasswordReinit'] = True
    def askPasswordReinit(self, tool):
        '''A user (anonymous) does not remember its password. Send him a mail
           containing a link that will trigger password re-initialisation.'''
        login = tool.req.rlogin
        _ = tool.translate
        if not login:
            return tool.goto(message=_('action_ko'))
        login = login.strip()
        user = tool.search1('User', login=login)
        message = _('reinit_mail_sent')
        if not user:
            # Return the message nevertheless. This way, malicious users can't
            # deduce information about existing users.
            return tool.goto(message=message)
        # If login is an email, use it. Else, use user.email instead.
        email = user.login
        if not String.EMAIL.match(email):
            email = user.email
        if not email:
            # Impossible to re-initialise the password
            return tool.goto(message=message)
        # Create a temporary file whose name is the user login and whose
        # content is a generated token.
        token = Password().generate()
        with (Path(putils.getOsTempFolder()) / login).open('w') as f:
            f.write(token)
        # Send an email
        initUrl = '%s/guard/doPasswordReinit?login=%s&token=%s' % \
                  (tool.url, login, token)
        subject = _('reinit_password')
        map = {'url':initUrl, 'siteUrl':tool.siteUrl}
        body= _('reinit_password_body', mapping=map, asText=True)
        tool.sendMail(email, subject, body)
        return tool.goto(message=message)

    traverse['doPasswordReinit'] = True
    def doPasswordReinit(self, tool):
        '''Performs the password re-initialisation'''
        req = tool.req
        login = req.login
        token = req.token
        _ = tool.translate
        siteUrl = tool.computeHomePage()
        # Stop if the URL was called with missing parameters
        if not login or not token:
            return tool.goto(siteUrl, _('wrong_password_reinit'))
        # Check if such token exists in temp folder
        tokenFile = Path(putils.getOsTempFolder()) / login
        if tokenFile.exists():
            with tokenFile.open() as f:
                storedToken = f.read()
                if storedToken == token:
                    # A commit is required
                    tool.H().commit = True
                    # Generate a new password for this user
                    user = tool.search1('User', login=login)
                    newPassword = Password().generate()
                    user.password = newPassword
                    # Send the new password by email
                    recipient = user.getMailRecipient()
                    subject = _('new_password')
                    map = {'password': newPassword, 'siteUrl': tool.siteUrl}
                    body = _('new_password_body', mapping=map, asText=True)
                    tool.sendMail(recipient, subject, body)
                    tokenFile.unlink()
                    return tool.goto(siteUrl, _('new_password_sent'))
        # If here, something went wrong
        return tool.goto(siteUrl, _('wrong_password_reinit'))

    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # PXs
    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    # A custom zone, at the bottom of pxLogin, allowing an app to add custom
    # content.
    pxLoginBottom = Px('')

    # The login form
    pxLogin = Px('''
     <!-- Popup for reinitializing the password -->
     <div if="isAnon and config.security.activateForgotPassword"
          id="askPasswordReinitPopup" class="popup">
      <form id="askPasswordReinitForm" method="post"
            action=":'%s/guard/askPasswordReinit' % tool.url">
       <div align="center">
        <p>:_('app_login')</p>
        <input type="text" size="35" name="rlogin" id="rlogin" value=""/>
        <br/><br/>
        <input type="button" onclick="doAskPasswordReinit()"
               value=":_('ask_password_reinit')"/>
        <input type="button" onclick="closePopup('askPasswordReinitPopup')"
               value=":_('object_cancel')"/>
       </div>
      </form>
     </div>

     <!-- The login box -->
     <div class="loginBox"
          var="cfg=config.security;
               discreet=config.ui.discreetLogin;
               display='none' if discreet and not req.stay else 'block'"
          style=":'display:%s' % display">

      <!-- Allow to hide the box when relevant -->
      <img if="discreet == True" src=":url('close.svg')" style="float:right"
           onclick="toggleLoginBox(false)" class="clickable iconXS"/>

      <div align="center"><img src=":url('loginLogo')"/></div>
      <h1>::_('app_name')</h1>
      <center class="sub">:_('please_connect')</center>
      <div class="loginSpace"></div>
      <form id="loginForm" name="loginForm" method="post" class="login"
            action=":'%s/guard/enter' % tool.url">

       <!-- Login fields  -->
       <center><input type="text" name="login" id="login" value=""
                   placeholder=":_('app_login')"/></center>
       <center><input type="password" name="password" id="password"
                placeholder=":_('app_password')"/></center>

       <!-- The authentication context -->
       <x var="ctx=cfg.authContext"
          if="ctx and ctx.chooseOnLogin">:ctx.pxOnLogin</x>

       <!-- Additional authentification data to carry -->
       <input type="hidden" name="authInfo" id="authInfo"
              value=":req.authInfo or ''"/>

       <!-- Must we go to some page after login ? -->
       <input type="hidden" name="goto" value=":req.get('goto', '')"/>

       <!-- The "submit" button -->
       <div class="submitLogin">
         <input type="submit" name="submit" var="label=_('app_connect')"
                value=":label" alt=":label"/></div>

       <!-- Forgot password ? -->
       <div class="lostPassword" if="cfg.activateForgotPassword">
         <a class="clickable"
         onclick="openPopup('askPasswordReinitPopup')">:_('forgot_password')</a>
       </div>

       <!-- Link to self-registration -->
       <div if="cfg.selfRegister" class="autoRegister">
        <span style="padding-right: 10px">:_('want_register')</span>
          <a style="font-size: 90%" class="clickable"
             onclick=":guard.getSelfRegisterJS(tool)">:_('app_register')</a>
       </div>
      </form>

      <!-- Custom zone -->
      <x>:guard.pxLoginBottom</x>
     </div>''',

     js='''
       // Function triggered when the user asks password reinitialisation
       function doAskPasswordReinit() {
         // Check that the user has typed a login
         var f = document.getElementById('askPasswordReinitForm'),
             login = f.rlogin.value.replace(' ', '');
         if (!login) { f.rlogin.style.background = wrongTextInput; }
         else {
           closePopup('askPasswordReinitPopup');
           f.submit();
         }
       }''',

     css='''
      .loginBox {
         border:|loginBorder| |loginColorPh|; padding:35px; color:|darkColor|;
         background-color:|boxBgColor|; position:absolute; top:|loginTop|;
         left:50%; z-index:10; transform:translate(-50%,-50%);
         box-shadow:|loginShadow|; border-radius: |loginBorderRadius|
      }
      .loginBox a, .loginBox a:visited { font-weight:bold; letter-spacing:1px }
      .loginBox h1 { padding-bottom:0; margin-bottom:3px; text-align:center;
                     font-size:160%; font-weight: |loginTitleWeight| }
      .loginBox .sub { font-size:120% }
      #loginForm input {
        background-color:|loginBgColor|; width:|loginWidth|; border:none;
        padding:|loginPadding|; margin:|loginMargin|; color:|loginColor| }
      .submitLogin { text-align:|submitAlign|
      }
      #loginForm input[type=submit] {
        color:|submitColor|; width:|submitWidth|; height:|submitHeight|;
        background:|submitBgColor| |submitBgUrl| no-repeat center left;
        border-radius:|submitBgRadius|; font-weight:|submitWeight|;
        margin-top:|submitTop|; text-transform:uppercase;
        padding:|submitPadding|
      }
      .lostPassword { padding:10px 0; font-size:80%; text-align:|loginAlign| }
      .autoRegister { text-align:|loginAlign| }

      /* Placeholder color for the login form */
      .loginBox ::placeholder, .loginBox ::-webkit-input-placeholder {
        color:|loginColorPh|; font-weight:bold; text-transform:uppercase }
      .loginSpace { height:35px }
      #rlogin { color:white }
     ''')

    # Warning about using cookies  - - - - - - - - - - - - - - - - - - - - - - -

    pxWarnCookies = Px('''
     <div class="cookieWarn">
      <div>
       <x>:_('cookie_warn')</x>
       <a href=":getattr(tool, config.security.cookieWarning)">:_('cookie_read')
       </a>
       <input type="button" value=":_('cookie_accept')"
              onclick="setAcceptCookie(this)"/>
      </div>
     </div>
    ''',

    css='''
      .cookieWarn { background-color:lightgrey; width:100%; position:absolute;
                    bottom:0; font-size:90%; z-index: 100 }
      .cookieWarn div { padding: 7px 15px }
      .cookieWarn a { margin: 0 10px; text-decoration: underline }
      .cookieWarn input { display:inline; background-color:#b5b5b5 }
    ''',

    js='''
      function setAcceptCookie(button) {
        createCookie('cookiesAccepted', '1');
        button.parentNode.parentNode.style.display = 'none';
      }'''
    )
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



'''HTTP communication between peer Appy sites'''

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
import time, random

from appy.peer import importer
from appy.utils.client import Resource
from appy.model.utils import Object as O
from appy.model.utils import notFromPortlet
from appy.peer.unresolved import Unresolved

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
IMPORT_START = 'Importing data from Appy 0 site %s...'
IMPORT_END   = "Done in %.2f''."
IMPORT_KO    = 'Import aborted.'
IMPORT_404   = '404 on %s.'
ROOT_START   = 'Importing instance(s) of root class "%s"...'
ROOT_RUN     = ' > Reading %d distant "%s" instances...'
ROOT_END     = '%d instance(s) of root class "%s" imported.'

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Config:
    '''Configuration about peer Appy sites'''

    def __init__(self):
        # Place here an instance of appy.peer.Appy0 if you want to initialise
        # this site with data migrated from a Appy 0.x site.
        self.appy0 = None

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Peer:
    '''Represents a peer Appy site to/from which to communicate, generally via
       HTTP.'''

    # Exception classes
    class NotFound(Exception): pass # 404 error
    class Error(Exception): pass # Any other error code

    # Peer sub-classes may declare an alternate base object importer
    baseImporter = importer.Importer

    # The ID of the tool, in Appy 1, is 'tool'
    distantToolName = 'tool'

    # Symbols used for script output
    printSymbols = ('-', '-', '-', '°', '~', ':')

    # Peer constructor - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    def __init__(self, url, login=None, password=None, id=None, name=None,
                 myId=None, myName=None, folder=None):
        # The URL of the distant site
        self.url = url
        # An optional ID for this distant site
        self.id = id or url
        # An optional human-readable name for the distant site
        self.name = name or self.id
        # The ID of the site running this code, as known by peer sites
        self.myId = myId
        # A human-readable name for the site running this code
        self.myName = myName or myId
        # It may be important to know the absolute disk path to the distant site
        # (as a pathlib.Path instance).
        self.folder = folder
        # Attribute "importers" below allows to define specific importer classes
        # for specific Appy classes. When importing some instance, if its local
        # class is mentioned in "importers", it is used. Else, the base importer
        # is used, as set on the Peer's static attribute "baseImporter" (that,
        # itself, may be overridden by a Peer sub-class).
        self.importers = {'User': importer.UserImporter}
        # When importing objects from one site to another, attribute "add" on
        # Ref fields is used to determine if tied objects must be retrieved
        # distantly and localy created (add=True) or simply found and linked
        # (add=False). That being said, it may happen, in the code of an app,
        # that a Ref field is declared to be "add=False,link=False", because
        # every action on the Ref is done programmatically. In that case, if the
        # Ref is composite=True, it will nevertheless be considered as addable.
        # If not, it is not possible to automatically deduce its "addability".
        # If your app uses Refs being "add=False,link=False,composite=False" but
        # are "addable", you must declare them in the following attribute. If
        # you don't do it, those Refs will be considered as not addable.
        self.addable = {} # ~{s_className: {s_refFieldName: None}}~
        # ~~~
        # The following attributes store run-time temporary data structures
        # ~~~
        # Create the proxy instance representing the distant server
        self.server = Resource(url, login, password)
        # A temporary dict of local objects, keyed by their IDs on the distant
        # site. This is required to reify links between local objects via Refs.
        self.localObjects = {}
        # Unresolved links between objects (for now on)
        self.unresolved = Unresolved(self)
        # If you let the following attribute empty, instances of all classes
        # considered as root classes will be imported from the distant site. A
        # class is considered being "root" if its name appears in config
        # attribute "config.model.rootClasses" AND if class attribute's
        # "createVia" is not set to standard method
        #
        #                  appy.model.utils.notFromPortlet
        #
        # Indeed, root classes defined with createVia=notFromPortlet are not
        # real root classes: they have been added in config.model.rootClasses
        # to benefit from the portlet's search facilities. If your app does not
        # use method "notFromPortlet" or if you want to tailor the import
        # process, you can bypass this logic and explicitly specify, in the
        # following attribute, names of the root classes for which instances
        # must be imported. Beware: it can lead to problems if, due to this
        # customization, some instances are not imported at all but are
        # mentioned in Refs from other imported objects.
        self.root = None
        # If the names of the distant classes differ from the names of the
        # classes as defined on this site's app, you may specify a name mapping
        # in the following attribute. Any class not found in this dict will be
        # considered to have the same name on both sites.
        self.classNames = {}

    def get(self, url=None, params=None, post=False):
        '''Sends a HTTP GET request (or POST if p_post is True) to p_url
           (or p_self.url if not given) with p_params.'''
        try:
            if post:
                r = self.server.post(uri=url, data=params)
            else:
                r = self.server.get(uri=url, params=params)
        except Resource.Error as re:
            raise self.Error('%s: %s' % (url or self.url, str(re)))
        if r.code == 404:
            raise self.NotFound(r)
        elif r.code != 200:
            raise self.Error(r)
        return r.data

    def getImporter(self, class_):
        '''Returns the importer class being suitable for importing instances of
           this p_class_.'''
        return self.importers.get(class_.name) or self.baseImporter

    def getRootClasses(self, tool):
        '''Returns the names of the root classes for which instances must be
           imported.'''
        # Root classes may be explicitly specified on this Peer instance
        if self.root: return self.root
        r = []
        for class_ in tool.model.getRootClasses():
            # Ignore classes with createVia=notFromPortlet
            if getattr(class_.python, 'createVia', None) == notFromPortlet:
                continue
            r.append(class_.name)
        return r

    def getIdFromUrl(self, url):
        '''Extract the ID of an object from its URL'''
        r = url.rsplit('/', 2)[-2]
        return int(r) if r.isdigit() else r

    def getSearchUrl(self, tool, className):
        '''Return the URL allowing to retrieve URLs of all instances from a
           given p_className on a distant site.'''
        # Get the name of the distant class corresponding to this local
        # p_className.
        distantName = self.classNames.get(className) or className
        return '%s/searchAll?className=%s' % (tool.url, distantName)

    def printSymbol(self, nb):
        '''Don't get bored while observing script execution'''
        if (nb % 100) == 0:
            # Every 100 elements, show the current element's p_nb
            r = '-=%d=-' % nb
        else:
            # Print a random char
            r = random.choice(self.printSymbols)
        print(r, end='', flush=True)

    def pumpClass(self, tool, className):
        '''Imports all instances from p_className'''
        tool.log(ROOT_START % className)
        # Retrieve the base URL allowing to retrieve URLs of instances of
        # p_className on the distant site.
        searchUrl = self.getSearchUrl(tool, className)
        urls = self.get(searchUrl)
        class_ = tool.model.classes[className]
        importer = self.getImporter(class_)
        counts = O(current=0, created=0)
        tool.log(ROOT_RUN % (len(urls), className))
        for url in urls:
            counts.current += 1
            self.printSymbol(counts.current)
            # Get the distant object and its ID
            distant = self.get(url)
            distantId = self.getIdFromUrl(url)
            # Must we really import this object ?
            action, detail = importer.getAction(distant)
            if action != importer.CREATE: continue
            # Create a local object corresponding to the distant one
            o = tool.create(className, executeMethods=False)
            # Fill local object's attributes from distant ones
            importer(self, o, distant).run()
            # Add it to the dict of already imported objects
            self.localObjects[distantId] = o
            # Count it
            counts.created += 1
        tool.log(ROOT_END % (counts.created, className))

    def pump(self, tool):
        '''Imports data from a distant Appy 0 site'''
        start = time.time()
        url = self.url
        tool.log(IMPORT_START % url)
        # The starting point of the whole importing process is the tool
        try:
            url = '%s/%s/xml' % (url, self.distantToolName)
            dtool = self.get(url)
            importer = self.getImporter(tool.class_)
            importer(self, tool, dtool,
                     importHistory=False, importLocalRoles=False).run()
            # After the tool and its dependent objects, import root classes and
            # their dependent objects.
            for className in self.getRootClasses(tool):
                self.pumpClass(tool, className)
        except Peer.Error as err:
            tool.log(str(err))
            tool.log(IMPORT_KO)
        except Peer.NotFound as err:
            tool.log(IMPORT_404 % err)
        else:
            # Resolve missing links
            self.unresolved.resolve(tool)
            tool.log(IMPORT_END % (time.time() - start))
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
from appy.xml import Parser
from appy.utils.string import Normalize

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
EXTRACTOR_WRONG_PARAMS = "You can't have both 'keepCRs' and 'normalize' " \
  "being True. Normalizing implies CRs to be removed."

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Extractor(Parser):
    '''Produces a text version of XML/XHTML content'''
    # There are 2 main usages to this parser. It can extract text:
    # - to produce a complete but textual version of some XHTML chunk, keeping
    #   accented chars and carriage returns;
    # - to extract text for search purposes, converting any accented char to its
    #   non-accented counterpart and removing any other element.

    paraTags = ('p', 'li', 'center', 'div', 'blockquote')

    def __init__(self, keepCRs=True, normalize=False, lower=True,
                       keepDashes=False, raiseOnError=False):
        Parser.__init__(self, raiseOnError=raiseOnError)
        # Must we keep carriage returns (and thus keep the global splitting of
        # the text into paragraphs) ?
        self.keepCRs = keepCRs
        # Must text be normalized ? When True, every accented char is converted
        # to its non-accented counterpart.
        self.normalize = normalize
        # Is is not possible to have both p_keepCRs and p_normalize being True
        if keepCRs and normalize: raise Exception(EXTRACTOR_WRONG_PARAMS)
        # Must be lowerise text ? (only if p_normalize is True)
        self.lower = lower
        # Must we keep dashes ? (only if p_normalize is True)
        self.keepDashes = keepDashes

    def startDocument(self):
        Parser.startDocument(self)
        self.r = []

    def endDocument(self):
        sep = '' if self.keepCRs else ' '
        self.r = sep.join(self.r)
        return Parser.endDocument(self)

    def characters(self, content):
        if self.normalize:
            content = Normalize.text(content, lower=self.lower,
                                     keepDash=self.keepDashes)
            if len(content) <= 1: return
        else:
            # Even if we must keep CRs, those encountered here are not
            # significant.
            content = content.replace('\n', ' ')
        self.r.append(content)

    def startElement(self, name, attrs):
        '''In "non-normalizing" mode, dumps a carriage return every time a "br"
           tag is encountered'''
        if self.keepCRs and (name == 'br'): self.r.append('\n')

    def endElement(self, name):
        '''In "non-normalizing" mode, dumps a carriage return every time a
           paragraph is encountered.'''
        if self.keepCRs and (name in Extractor.paraTags): self.r.append('\n')
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------
from appy.fields import Field, Layouts

# ------------------------------------------------------------------------------
class Info(Field):
    '''An info is a field whose purpose is to present information
       (text, html...) to the user.'''
    # An info only displays a label. So PX for showing content are empty.
    pxView = pxEdit = pxCell = pxSearch = ''

    def __init__(self, validator=None, multiplicity=(1,1), show='view',
      page='main', group=None, layouts=None, move=0,
      specificReadPermission=False, specificWritePermission=False, width=None,
      height=None, maxChars=None, colspan=1, master=None, masterValue=None,
      focus=False, historized=False, mapping=None, generateLabel=None,
      label=None, view=None, cell=None, edit=None, xml=None, translations=None):
        Field.__init__(self, None, (0,1), None, None, show, page, group,
          layouts, move, False, True, None, False, specificReadPermission,
          specificWritePermission, width, height, None, colspan, master,
          masterValue, focus, historized, mapping, generateLabel, label, None,
          None, None, None, False, False, view, cell, edit, xml, translations)
        self.validable = False

    def getDefaultLayouts(self): return Layouts.Info.b
# ------------------------------------------------------------------------------



'''Object's field "title" requires a specific treatment in order to be rendered
   in the user interface.'''

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
from appy.px import Px
from appy.ui.criteria import Criteria

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Title:
    '''Manages the zone, mainly within lists, rendering the title of an object
       and everything around it (sup- and sub-titles).'''

    @staticmethod
    def showSub(class_, field):
        '''Show button "details" only under field "title", if a method named
           "getSubTitle" is defined on p_class_.'''
        return (field.name == 'title') and class_.hasSubTitle()

    # Icon for hiding/showing the "sub-title" zone under an object's title
    pxSub = Px('''
     <img src=":url('toggleDetails')" class="clickable"
          onclick="toggleSubTitles()"/>''',
    
     js='''
      // Function that sets a value for showing/hiding sub-titles
      setSubTitles = function(value, tag) {
        createCookie('showSubTitles', value);
        // Get the sub-titles
        var subTitles = getElementsHavingName(tag, 'subTitle');
        if (subTitles.length == 0) return;
        // Define the display style depending on p_tag
        var displayStyle = 'inline';
        if (tag == 'tr') displayStyle = 'table-row';
        for (var i=0; i < subTitles.length; i++) {
          if (value == 'True') subTitles[i].style.display = displayStyle;
          else subTitles[i].style.display = 'none';
        }
      }

      // Function that toggles the value for showing/hiding sub-titles
      toggleSubTitles = function(tag) {
        // Get the current value
        var value = readCookie('showSubTitles') || 'True';
        // Toggle the value
        var newValue = (value == 'True')? 'False': 'True';
        if (!tag) tag = 'div';
        setSubTitles(newValue, tag);
      }''')

    @classmethod
    def get(class_, o, mode='link', nav='no', target=None, page='main',
            popup=False, baseUrl=None, title=None, linkTitle=None, css=None,
            selectJs=None, highlight=False, backHook=None, maxChars=None,
            pageLayout=None):
        '''Gets p_o's title as it must appear in lists of objects (ie in
           lists of tied objects in a Ref, in search results...).'''

        # In most cases, a title must appear as a link that leads to the object
        # view layout. In this case (p_mode == "link"):
        # ----------------------------------------------------------------------
        # p_nav        | is the navigation parameter allowing navigation between
        #              | this object and others;
        # p_target     | specifies if the link must be opened in the popup or
        #              | not. It is an instance of appy.ui.LinkTarget;
        # p_page       | specifies which page to show on the target object view;
        # p_popup      | indicates if we are already in the popup or not;
        # p_baseUrl    | indicates a possible alternate base URL for accessing
        #              | the object;
        # p_title      | specifies, if given, an alternate content for the "a"
        #              | tag (can be a PX);
        # p_linkTitle  | specifies a possible value for attribute "link" for the
        #              | "a" tag (by default this attribute is not dumped);
        # p_css        | can be the name of a CSS class to apply (also for other
        #              | modes);
        # p_maxChars   | gives an (optional) limit to the number of visible
        #              | title chars.
        # p_pageLayout | allows to override the standard "view" page layout as
        #              | defined on p_o's class.
        # ----------------------------------------------------------------------

        # Another p_mode is "select". In this case, we are in a popup for
        # selecting objects: every title must not be a link, but clicking on it
        # must trigger Javascript code (in p_selectJs) that will select this
        # object.

        # The last p_mode is "text". In this case, we simply show the object
        # title but with no tied action (link, select).

        # If p_highlight is True, keywords will be highlighted if we are in the
        # context of a query with keywords.

        # If the class_ whose elements must be listed has an attribute
        # "uniqueLinks" being True, in "link" mode, the generated link will
        # include an additional parameter named "_hash". This parameter will
        # ensure that the link will be different every time it is generated.
        # This is useful for short-circuiting the a:visited CSS style when an
        # app wants to manage link's style in some specific way.

        # Compute CSS class
        class_ = o.class_
        handler = o.H()
        cssClass = class_.getCssFor(o, 'title')
        if css: cssClass += ' %s' % css
        # Get the title, with highlighted parts when relevant
        titleIsPx = False
        if not title:
            title = class_.getListTitle(o)
        elif isinstance(title, Px):
            title = title(handler.context)
            titleIsPx = True
        if highlight and not titleIsPx:
            title = Criteria.highlight(handler, title)
        if maxChars: title = Px.truncateText(title, width=maxChars)
        if mode.endswith('link'): # "link" or "dlink"
            popup = popup or (target.target != '_self')
            params = {'page': page, 'nav': nav, 'popup': popup,
                      'unique': class_.produceUniqueLinks()}
            if pageLayout: params['pageLayout'] = pageLayout
            # Build the link URL, or patch the given p_baseUrl
            url = o.getUrl(sub='view', **params) \
                  if baseUrl is None else o.patchUrl(baseUrl, **params)
            # Define attribute "onClick"
            if target.onClick:
                onClick = ' onclick="%s"' % \
                          target.getOnClick(backHook or str(o.iid), o)
            else:
                onClick = ''
            # Set a "title" parameter when relevant
            lt = linkTitle and (' title="%s"' % linkTitle) or ''
            if mode[0] == 'd':
                # Wrap the link into a "div". In that case, the CSS class
                # applies to the div.
                start = '<div class="%s"><a' % cssClass
                end = '</a></div>'
            else:
                start = '<a class="%s"' % cssClass
                end = '</a>'
            r = '%s name="title" href="%s" target="%s"%s%s>%s%s' %\
                (start, url, target.target, onClick, lt, title, end)
        elif mode == 'select':
            r = '<span class="%s clickable" onclick="%s">%s</span>' % \
                (cssClass, selectJs, title)
        elif mode == 'text':
            r = '<span class="%s">%s</span>' % (cssClass, title)
        return r
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



# -*- coding: utf-8 -*-
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------
import re, random, sys
from appy.px import Px
from appy.fields import Field
from appy.gen.layout import Table
from appy.shared.diff import HtmlDiff
from appy.shared.data import countries
from appy.shared import utils as sutils
from appy.fields.select import Selection
from appy.shared.xml_parser import Escape
from appy.shared.xml_parser import XhtmlCleaner
from appy.gen.indexer import XhtmlTextExtractor, ValidValue

# ------------------------------------------------------------------------------
digit  = re.compile('[0-9]')
alpha  = re.compile('[a-zA-Z0-9]')
letter = re.compile('[a-zA-Z]')
digits = '0123456789'
letters = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ'
emptyTuple = ()

# ------------------------------------------------------------------------------
NOT_IMPLEMENTED = 'This feature is not implemented yet'
WRONG_READONLY = 'A read-only field must have a format being LINE.'
SELECT_MULTILINGUAL = "A selection field can't be multilingual."
PASSWORD_MULTILINGUAL = "A password or captcha field can't be multilingual."

# ------------------------------------------------------------------------------
class String(Field):
    # Some base layouts for Strings
    tLayouts = {'view': 'l-f', 'edit': 'lrv-d-f'} # For TEXT format
    gtLayouts = {'view': Table('fl', width='99%'), # Idem but in a grid group
                 'edit': Table('d2-f;rv=', width='99%')}
    xLayouts = tLayouts # For XHTML format
    xcLayouts = {'view': 'lc-f', 'edit': 'lrv-d-f'} # Idem but with history
    gxLayouts = gtLayouts # Idem but in a grid group
    gxcLayouts = {'view': Table('cl-f', width='99%'),
                  'edit': Table('drv-f', width='99%')} # Idem but with history
    stLayouts = {'view': Table('fl', width=None), # For select strings
                 'edit': Table('d2-f;rv=', width=None)}

    # Javascript files sometimes required by this type. Method String.getJs
    # below determines when the files must be included.
    cdnUrl = '//cdn.ckeditor.com/%s/%s/ckeditor.js'

    # Default custom style, allowing to produce, in POD results, a page break
    # just after the paragraph having this style.
    customStyles = [{'id': 'div', 'element': 'div',
                     'attributes': {'style': 'page-break-after:always'}}]

    # Use this constant to say that there is no maximum size for a string field
    NO_MAX = sys.maxint

    # Some predefined regular expressions that may be used as validators
    c = re.compile
    EMAIL = c('[a-zA-Z][\w\.-]*[a-zA-Z0-9]*@[a-zA-Z0-9][\w\.-]*[a-zA-Z0-9]\.' \
              '[a-zA-Z][a-zA-Z\.]*[a-zA-Z]')
    ALPHANUMERIC = c('[\w-]+')
    URL = c('(http|https):\/\/[a-z0-9]+([\-\.]{1}[a-z0-9]+)*(\.[a-z]{2,5})?' \
            '(([0-9]{1,5})?\/.*)?')

    # Possible values for "format"
    LINE = 0
    TEXT = 1
    XHTML = 2
    PASSWORD = 3
    CAPTCHA = 4

    # Default ways to render multilingual fields
    defaultLanguagesLayouts = {
      LINE:  {'edit': 'vertical',   'view': 'vertical', 'cell': 'vertical'},
      TEXT:  {'edit': 'horizontal', 'view': 'vertical', 'cell': 'vertical'},
      XHTML: {'edit': 'horizontal', 'view': 'horizontal', 'cell': 'vertical'},
    }

    # Python string methods to use to apply a transform
    transformMethods = {'uppercase': 'upper', 'lowercase': 'lower',
                        'capitalize': 'capitalize'}

    # pxView part for formats String.LINE (but that are not selections) and
    # String.PASSWORD (a fake view for String.PASSWORD and no view at all for
    # String.CAPTCHA).
    pxViewLine = Px('''
     <span if="not value" class="smaller">-</span>
     <x if="value">
      <!-- A password -->
      <x if="fmt == 3">********</x>
      <!-- A URL -->
      <a if="(fmt != 3) and isUrl" target="_blank" href=":value">:value</a>
      <!-- Any other value -->
      <x if="(fmt != 3) and not isUrl">::value</x>
     </x>''')

    # pxView part for format String.TEXT
    pxViewText = Px('''
     <span if="not value" class="smaller">-</span><x if="value">::value</x>''')

    # pxView part for format String.XHTML
    pxViewRich = Px('''
     <div if="not mayAjaxEdit" class="xhtml">::value or '-'</div>
     <x if="mayAjaxEdit" var2="name=lg and ('%s_%s' % (name, lg)) or name">
      <div class="xhtml" contenteditable="true"
           id=":'%s_%s_ck' % (zobj.id, name)">::value or '-'</div>
      <script if="mayAjaxEdit">::field.getJsInlineInit(zobj, name, lg)</script>
     </x>''')

    # PX displaying the language code and name besides the part of the
    # multilingual field storing content in this language.
    pxLanguage = Px('''
     <td style=":'padding-top:%dpx' % lgTop" width="25px">
      <span class="language help"
            title=":ztool.getLanguageName(lg)">:lg.upper()</span>
     </td>''')

    pxMultilingual = Px('''
     <!-- Horizontally-layouted multilingual field -->
     <table if="mLayout == 'horizontal'" width="100%"
            class=":(layoutType == 'cell') and 'no' or ''"
            var="count=len(languages)">
      <tr valign="top"><x for="lg in languages"><x>:field.pxLanguage</x>
       <td width=":'%d%%' % int(100.0/count)"
           var="requestValue=requestValue[lg]|None;
                value=value[lg]|emptyDefault">:field.subPx[layoutType][fmt]</td>
      </x></tr></table>

     <!-- Vertically-layouted multilingual field -->
     <table if="mLayout == 'vertical'"
            class=":(layoutType == 'cell') and 'no' or ''">
      <tr valign="top" height="20px" for="lg in languages">
       <x>:field.pxLanguage</x>
       <td var="requestValue=requestValue[lg]|None;
                value=value[lg]|emptyDefault">:field.subPx[layoutType][fmt]</td>
     </tr></table>''')

    pxView = Px('''
     <x var="fmt=field.format; isUrl=field.isUrl;
             languages=field.getAttribute(obj, 'languages');
             multilingual=len(languages) &gt; 1;
             mLayout=multilingual and field.getLanguagesLayout('view');
             inlineEdit=field.getAttribute(obj, 'inlineEdit');
             mayAjaxEdit=(layoutType != 'cell') and not showChanges and \
                         inlineEdit and zobj.mayEdit(field.writePermission)">
      <x if="field.isSelect">
       <span if="not value" class="smaller">-</span>
       <x if="value and not isMultiple">::value</x>
       <ul if="value and isMultiple"><li for="sv in value"><i>::sv</i></li></ul>
      </x>
      <!-- Any other unilingual field -->
      <x if="not field.isSelect and not multilingual"
         var2="lg=None">:field.subPx['view'][fmt]</x>
      <!-- Any other multilingual field -->
      <x if="not field.isSelect and multilingual"
         var2="lgTop=1; emptyDefault='-'">:field.pxMultilingual</x>
      <!-- If this field is a master field -->
      <input type="hidden" if="masterCss" class=":masterCss" value=":rawValue"
             name=":name" id=":name"/></x>''')

    # pxEdit part for formats String.LINE (but that are not selections),
    # String.PASSWORD and String.CAPTCHA.
    pxEditLine = Px('''
     <input var="inputId=not lg and name or '%s_%s' % (name, lg);
                 placeholder=field.getPlaceholder(obj)"
            id=":inputId" name=":inputId" size=":field.getLineWidth()"
            maxlength=":field.maxChars" placeholder=":placeholder"
            value=":field.getInputValue(inRequest, requestValue, value)"
            style=":'text-transform:%s;%s' % \
                    (field.transform, field.getLineWidth(False))"
            type=":(fmt == 3) and 'password' or 'text'"
            readonly=":field.isReadonly(zobj)"/>
     <!-- Display a captcha if required -->
     <span if="fmt == 4">:_('captcha_text', \
                            mapping=field.getCaptchaChallenge(req.SESSION))
     </span>''')

    # pxEdit part for formats String.TEXT and String.XHTML
    pxEditTextArea = Px('''
     <textarea var="inputId=not lg and name or '%s_%s' % (name, lg)"
       id=":inputId" name=":inputId" cols=":field.getTextareaCols()"
       style=":field.getTextareaStyle()"
       rows=":field.height">:field.getInputValue(inRequest, requestValue, value)
     </textarea>
     <script if="fmt == 2">::field.getJsInit(zobj, lg)</script>''')

    pxEdit = Px('''
     <x var="fmt=field.format;
             languages=field.getAttribute(zobj, 'languages');
             multilingual=len(languages) &gt; 1;
             mLayout=multilingual and field.getLanguagesLayout('edit')">
      <select if="field.isSelect"
              var2="possibleValues=field.getPossibleValues(zobj, \
                      withTranslations=True, withBlankValue=True);
                    charsWidth=field.getWidthInChars(False)"
              name=":name" id=":name" class=":masterCss" multiple=":isMultiple"
              onchange=":field.getOnChange(zobj, layoutType)"
              size=":field.getSelectSize(False, isMultiple)"
              style=":field.getSelectStyle(False, isMultiple)">
       <option for="val in possibleValues" value=":val[0]"
               selected=":field.isSelected(zobj, name, val[0], rawValue)"
               title=":val[1]">:ztool.truncateValue(val[1], charsWidth)</option>
      </select>
      <!-- Any other unilingual field -->
      <x if="not field.isSelect and not multilingual"
         var2="lg=None">:field.subPx['edit'][fmt]</x>
      <!-- Any other multilingual field -->
      <x if="not field.isSelect and multilingual"
         var2="lgTop=(fmt!=2) and 3 or 1;
               emptyDefault=''">:field.pxMultilingual</x>
      </x>''')

    pxCell = Px('''
     <x if="field.isUrl" var2="value=field.getValueIf(zobj, name, layoutType)">
      <a href=":value" title=":value" target="_blank">
       <img src=":url('url')"/></a></x>
     <x if="not field.isUrl"
        var2="multipleValues=value and isMultiple">
      <x if="multipleValues">:', '.join(value)</x>
      <x if="not multipleValues">:field.pxView</x>
     </x>''')

    pxSearch = Px('''
     <!-- Show a simple search field for most String fields -->
     <input if="not field.isSelect" type="text" maxlength=":field.maxChars"
            size=":field.swidth" value=":field.sdefault" name=":widgetName"
            style=":'text-transform:%s' % field.transform"/>
     <!-- Show a multi-selection box for fields whose validator defines a list
          of values, with a "AND/OR" checkbox. -->
     <x if="field.isSelect">
      <!-- The "and" / "or" radio buttons -->
      <x if="field.multiplicity[1] != 1"
         var2="operName='o_%s' % name;
              orName='%s_or' % operName;
              andName='%s_and' % operName">
       <input type="radio" name=":operName" id=":orName" checked="checked"
              value="or"/>
       <label lfor=":orName">:_('search_or')</label>
       <input type="radio" name=":operName" id=":andName" value="and"/>
       <label lfor=":andName">:_('search_and')</label><br/>
      </x>
      <!-- The list of values -->
      <select var="preSelected=field.sdefault;
                   charsWidth=field.getWidthInChars(True)"
              name=":widgetName" multiple="multiple"
              size=":field.getSelectSize(True, True)"
              style=":field.getSelectStyle(True, True)"
              onchange=":field.getOnChange(ztool, 'search', className)">
       <option for="v in field.getPossibleValues(ztool, withTranslations=True,\
                                     withBlankValue=False, className=className)"
               selected=":v[0] in preSelected" value=":v[0]"
               title=":v[1]">:ztool.truncateValue(v[1], charsWidth)</option>
      </select>
     </x><br/>''')

    # Widget for filtering object values on query results
    pxFilterSelect = Px('''
     <select var="name=field.name;
                  filterId='%s_%s' % (mode.ajaxHookId, name);
                  charsWidth=field.getWidthInChars(True)"
        id=":filterId" name=":filterId" class="discreet"
          onchange=":'askBunchFiltered(%s,%s)' % (q(mode.ajaxHookId), q(name))">
      <option for="v in field.getPossibleValues(ztool, withTranslations=True,\
        withBlankValue='forced', blankLabel='everything', className=className)"
        selected=":(name in mode.filters) and (mode.filters[name] == v[0])"
        value=":v[0]" title=":v[1]">:ztool.truncateValue(v[1], \
                                                         charsWidth)</option>
     </select>''')

    # Sub-PX to use according to String format
    subPx = {
     'edit': {LINE:pxEditLine, TEXT:pxEditTextArea, XHTML:pxEditTextArea,
              PASSWORD:pxEditLine, CAPTCHA:pxEditLine},
     'view': {LINE:pxViewLine, TEXT:pxViewText, XHTML:pxViewRich,
              PASSWORD:pxViewLine, CAPTCHA:pxViewLine}
    }
    subPx['cell'] = subPx['view']

    # Some predefined functions that may also be used as validators
    @staticmethod
    def _MODULO_97(obj, value, complement=False):
        '''p_value must be a string representing a number, like a bank account.
           this function checks that the 2 last digits are the result of
           computing the modulo 97 of the previous digits. Any non-digit
           character is ignored. If p_complement is True, it does compute the
           complement of modulo 97 instead of modulo 97. p_obj is not used;
           it will be given by the Appy validation machinery, so it must be
           specified as parameter. The function returns True if the check is
           successful.'''
        if not value: return True
        # First, remove any non-digit char
        v = ''
        for c in value:
            if digit.match(c): v += c
        # There must be at least 3 digits for performing the check
        if len(v) < 3: return False
        # Separate the real number from the check digits
        number = int(v[:-2])
        checkNumber = int(v[-2:])
        # Perform the check
        if complement:
            return (97 - (number % 97)) == checkNumber
        else:
            # The check number can't be 0. In this case, we force it to be 97.
            # This is the way Belgian bank account numbers work. I hope this
            # behaviour is general enough to be implemented here.
            mod97 = (number % 97)
            if mod97 == 0: return checkNumber == 97
            else:          return checkNumber == mod97
    @staticmethod
    def MODULO_97(obj, value): return String._MODULO_97(obj, value)
    @staticmethod
    def MODULO_97_COMPLEMENT(obj, value):
        return String._MODULO_97(obj, value, True)
    BELGIAN_ENTERPRISE_NUMBER = MODULO_97_COMPLEMENT

    @staticmethod
    def BELGIAN_NISS(obj, value):
        '''Returns True if the NISS in p_value is valid'''
        if not value: return True
        # Remove any non-digit from nrn
        niss = sutils.keepDigits(value)
        # NISS must be made of 11 numbers
        if len(niss) != 11: return False
        # When NRN begins with 0 or 1, it must be prefixed with number "2" for
        # checking the modulo 97 complement.
        nissForModulo = niss
        if niss.startswith('0') or niss.startswith('1'):
            nissForModulo = '2'+niss
        # Check modulo 97 complement
        return String.MODULO_97_COMPLEMENT(None, nissForModulo)

    @staticmethod
    def IBAN(obj, value):
        '''Checks that p_value corresponds to a valid IBAN number. IBAN stands
           for International Bank Account Number (ISO 13616). If the number is
           valid, the method returns True.'''
        if not value: return True
        # First, remove any non-digit or non-letter char
        v = ''
        for c in value:
            if alpha.match(c): v += c
        # Maximum size is 34 chars
        if (len(v) < 8) or (len(v) > 34): return False
        # 2 first chars must be a valid country code
        if not countries.exists(v[:2].upper()): return False
        # 2 next chars are a control code whose value must be between 0 and 96.
        try:
            code = int(v[2:4])
            if (code < 0) or (code > 96): return False
        except ValueError:
            return False
        # Perform the checksum
        vv = v[4:] + v[:4] # Put the 4 first chars at the end.
        nv = ''
        for c in vv:
            # Convert each letter into a number (A=10, B=11, etc)
            # Ascii code for a is 65, so A=10 if we perform "minus 55"
            if letter.match(c): nv += str(ord(c.upper()) - 55)
            else: nv += c
        return int(nv) % 97 == 1

    @staticmethod
    def BIC(obj, value):
        '''Checks that p_value corresponds to a valid BIC number. BIC stands
           for Bank Identifier Code (ISO 9362). If the number is valid, the
           method returns True.'''
        if not value: return True
        # BIC number must be 8 or 11 chars
        if len(value) not in (8, 11): return False
        # 4 first chars, representing bank name, must be letters
        for c in value[:4]:
            if not letter.match(c): return False
        # 2 next chars must be a valid country code
        if not countries.exists(value[4:6].upper()): return False
        # Last chars represent some location within a country (a city, a
        # province...). They can only be letters or figures.
        for c in value[6:]:
            if not alpha.match(c): return False
        return True

    def __init__(self, validator=None, multiplicity=(0,1), default=None,
      defaultOnEdit=None, format=LINE, show=True, page='main', group=None,
      layouts=None, move=0, indexed=False, mustIndex=True, indexValue=None,
      searchable=False, specificReadPermission=False,
      specificWritePermission=False, width=None, height=None, maxChars=None,
      colspan=1, master=None, masterValue=None, focus=False, historized=False,
      mapping=None, generateLabel=None, label=None, sdefault='', scolspan=1,
      swidth=None, sheight=None, persist=True, transform='none',
      placeholder=None, styles=('p','h1','h2','h3','h4'), customStyles=None,
      allowImageUpload=False, spellcheck=False, languages=('en',),
      languagesLayouts=None, inlineEdit=False, toolbar='Standard', view=None,
      cell=None, edit=None, xml=None, translations=None,
      noValueLabel='choose_a_value', readonly=False):
        # According to format, the widget will be different: input field,
        # textarea, inline editor... Note that there can be only one String
        # field of format CAPTCHA by page, because the captcha challenge is
        # stored in the session at some global key.
        self.format = format
        self.isUrl = validator == String.URL
        # ~~~ The following attributes have sense only if format is XHTML ~~~
        # The list of styles that the user will be able to select in the styles
        # dropdown (within CKEditor) is defined hereafter.
        self.styles = styles
        # If you want to add custom, non-standard styles in the above-mentioned
        # dropdown, do not touch attribute "styles", but specify such entries in
        # the following attribute "customStyles". It must be a list or dict of
        # entries; every entry represents a custom style and must be a dict
        # having the following keys and values. Every key and value must be a
        # string.
        # ----------------------------------------------------------------------
        #    "id"    | An identifier for your style, that must be different from
        #            | any standard CKEditor style like h1, h2, p, div, etc;
        # ----------------------------------------------------------------------
        #   "name"   | A translated name for your style, that will appear in the
        #            | dropdown. Do not use any special (ie, accentuated) char
        #            | in the value: prefer the use of an HTML entity for
        #            | defining such char.
        # ----------------------------------------------------------------------
        #  "element" | The HTML tag that will surround the text onto which the
        #            | style will be applied. Do not use "blockquote", it does
        #            | not work. Non-standard tag "address" works well, or any
        #            | standard tag like h1, h2, etc. Note that if you use a
        #            | standard tag, it will work, but if you have also
        #            | activated it in attribute "style" hereabove, you won't be
        #            | able to make the difference in the result between both
        #            | styles because they will produce a similar result.
        # ----------------------------------------------------------------------
        self.customStyles = customStyles or String.customStyles
        # Do we allow the user to upload images in it ?
        self.allowImageUpload = allowImageUpload
        if allowImageUpload: raise Exception(NOT_IMPLEMENTED)
        # Do we run the CK spellchecker ?
        self.spellcheck = spellcheck
        # What toolbar is used ? Possible values are "Standard" or "Simple"
        self.toolbar = toolbar
        # "placeholder", similar to the HTML attribute of the same name, allows
        # to specify a short hint that describes the expected value of the input
        # field. It is shown inside the input field and disappears as soon as
        # the user encodes something in it. Works only for strings whose format
        # is LINE. Does not work with IE < 10. You can specify a method here,
        # that can, for example, return an internationalized value.
        self.placeholder = placeholder
        # If "languages" holds more than one language, the field will be
        # multi-lingual and several widgets will allow to edit/visualize the
        # field content in all the supported languages. The field is also used
        # by the CK spell checker.
        self.languages = languages
        # When content exists in several languages, how to render them? Either
        # horizontally (one below the other), or vertically (one besides the
        # other). Specify here a dict whose keys are layouts ("edit", "view")
        # and whose values are either "horizontal" or "vertical".
        self.languagesLayouts = languagesLayouts
        # ~~~ End of XHTML-related attributes ~~~
        # "transform" below has a direct impact on the text entered by the user.
        # It applies a transformation on it, exactly as does the CSS
        # "text-transform" property. Allowed values are those allowed for the
        # CSS property: "none" (default), "uppercase", "capitalize" or
        # "lowercase".
        self.transform = transform
        # When selecting a value from a "select" widget, the entry representing
        # no value is translated according to this label. The default one is
        # something like "[ choose ]", but if you prefer a less verbose version,
        # you can use "no_value" that simply displays a dash, or your own label.
        self.noValueLabel = noValueLabel
        # If attribute "readonly" is True (or stores a method returning True),
        # the rendered input field, on edit layouts, will have attribute
        # "readonly" set.
        self.readonly = readonly
        # Call the base constructor
        Field.__init__(self, validator, multiplicity, default, defaultOnEdit,
          show, page, group, layouts, move, indexed, mustIndex, indexValue,
          searchable, specificReadPermission, specificWritePermission, width,
          height, maxChars, colspan, master, masterValue, focus, historized,
          mapping, generateLabel, label, sdefault, scolspan, swidth, sheight,
          persist, inlineEdit, view, cell, edit, xml, translations)
        self.isSelect = self.isSelection()
        # If self.isSelect, self.sdefault must be a list of value(s)
        if self.isSelect and not sdefault:
            self.sdefault = []
        # Default width, height and maxChars vary according to String format
        if width is None:
            if format == String.TEXT:    self.width  = 60
            # This width corresponds to the standard width of an Appy page
            elif format == String.XHTML: self.width  = None
            else:                        self.width  = 30
        if height is None:
            if format == String.TEXT:    self.height = 5
            elif format == String.XHTML: self.height  = None
            elif self.isSelect:          self.height = 4
            else:                        self.height = 1
        if maxChars is None:
            if self.isSelect: pass
            elif format == String.LINE: self.maxChars = 256
            elif format == String.TEXT: self.maxChars = String.NO_MAX
            elif format == String.XHTML: self.maxChars = String.NO_MAX
            elif format == String.PASSWORD: self.maxChars = 20
        if self.indexed and (self.format in (String.LINE, String.TEXT)):
            self.filterPx = self.isSelect and 'pxFilterSelect' or 'pxFilterText'
        self.swidth = self.swidth or self.width
        self.sheight = self.sheight or self.height
        self.checkParameters()

    def checkParameters(self):
        '''Ensures this String is correctly defined'''
        error = None
        if self.isMultilingual(None):
            if self.isSelect:
                error = SELECT_MULTILINGUAL
            elif self.format in (String.PASSWORD, String.CAPTCHA):
                error = PASSWORD_MULTILINGUAL
        if self.readonly and (self.format != String.LINE):
            error = WRONG_READONLY
        if error: raise Exception(error)

    def isSelection(self):
        '''Does the validator of this type definition define a list of values
           into which the user must select one or more values?'''
        res = True
        if type(self.validator) in (list, tuple):
            for elem in self.validator:
                if not isinstance(elem, basestring):
                    res = False
                    break
        else:
            if not isinstance(self.validator, Selection):
                res = False
        return res

    def isRenderable(self, layoutType):
        '''A value being an URL is potentially renderable everywhere'''
        if self.isUrl: return True
        return layoutType != 'buttons'

    def isSortable(self, usage):
        '''Can this field be sortable ?'''
        if self.name == 'state': return
        if usage == 'search':
            r = self.indexed and not self.isMultiValued() and \
                not self.isSelection()
        elif usage == 'ref':
            r = (self.format == 0) and not self.isMultilingual(None, True)
        return r

    def isMultilingual(self, obj, dontKnow=False):
        '''Is this field multilingual ? If we don't know, say p_dontKnow.'''
        # In the following case, impossible to know: we say no.
        if not obj:
            if callable(self.languages): return dontKnow
            else: return len(self.languages) > 1
        return len(self.getAttribute(obj, 'languages')) > 1

    def getLineWidth(self, sizeAttr=True):
        '''When p_self is rendered as an input field of type "text", this method
           returns its width as must be defined, either in the "size" attribute,
           (p_sizeAttr is True) or as a CSS "width" attribute (p_sizeAttr is
           False).'''
        # If the field width is expressed as a percentage, the width must be
        # specified via the CSS "width" attribute. In all other cases, it is
        # specified via the "size" tag attribute.
        width = self.width
        isPercentage = width and isinstance(width, str) and width.endswith('%')
        if sizeAttr: # tag attribute "size"
            if isPercentage: return ''
            else: return width
        else: # CSS attribute "width"
            return isPercentage and ('width:%s' % width) or ''

    def getDefaultLayouts(self):
        '''Returns the default layouts for this type. Default layouts can vary
           acccording to format, multiplicity, group or history.'''
        # Is this field in a grid-style group ?
        inGrid = self.inGrid()
        # Defaults layouts depend on format
        if self.format == String.TEXT:
            return inGrid and String.gtLayouts or String.tLayouts
        elif self.format == String.XHTML:
            if self.historized:
                # self.historized can be a method or a boolean. If it is a
                # method, it means that under some condition, historization will
                # be enabled. So we come here also in this case.
                return inGrid and String.gxcLayouts or String.xcLayouts
            else:
                return inGrid and String.gxLayouts or String.xLayouts
        elif self.isSelection() and inGrid:
            return String.stLayouts

    def getLanguagesLayout(self, layoutType):
        '''Gets the way to render a multilingual field on p_layoutType'''
        if self.languagesLayouts and (layoutType in self.languagesLayouts):
            return self.languagesLayouts[layoutType]
        # Else, return a default value that depends of the format.
        return String.defaultLanguagesLayouts[self.format][layoutType]

    def getValue(self, obj, name=None, layout=None, noListIfSingleObj=False):
        # Cheat if this field represents p_obj's state
        if self.name == 'state': return obj.State()
        value = Field.getValue(self, obj, name, layout)
        if not value:
            if self.isMultiValued(): return emptyTuple
            else: return value
        if isinstance(value, basestring) and self.isMultiValued():
            value = [value]
        elif isinstance(value, tuple):
            value = list(value)
        return value

    def getCopyValue(self, obj):
        '''If the value is mutable (ie, a dict for a multilingual field), return
           a copy of it instead of the value stored in the database.'''
        res = self.getValue(obj)
        if isinstance(res, dict): res = res.copy()
        return res

    def valueIsInRequest(self, obj, request, name=None, layoutType='view'):
        # If we are on the search layout, p_obj, if not None, is certainly not
        # the p_obj we want here (can be a home object).
        if layoutType == 'search':
            return Field.valueIsInRequest(self, obj, request, name, layoutType)
        languages = self.getAttribute(obj, 'languages')
        if len(languages) == 1:
            return Field.valueIsInRequest(self, obj, request, name, layoutType)
        # Is is sufficient to check that at least one of the language-specific
        # values is in the request.
        return request.has_key('%s_%s' % (name, languages[0]))

    def getRequestValue(self, obj, requestName=None):
        '''The request value may be multilingual'''
        request = obj.REQUEST
        name = requestName or self.name
        languages = self.getAttribute(obj, 'languages')
        # A unilingual field
        if len(languages) == 1: return request.get(name, None)
        # A multilingual field
        res = {}
        for language in languages:
            res[language] = request.get('%s_%s' % (name, language), None)
        return res

    def isEmptyValue(self, obj, value):
        '''Returns True if the p_value must be considered as an empty value'''
        if not isinstance(value, dict):
            return Field.isEmptyValue(self, obj, value)
        # p_value is a dict of multilingual values. For such values, as soon
        # as a value is not empty for a given language, the whole value is
        # considered as not being empty.
        for v in value.itervalues():
            if not Field.isEmptyValue(self, obj, v): return
        return True

    def isCompleteValue(self, obj, value):
        '''Returns True if the p_value must be considered as complete. For a
           unilingual field, being complete simply means not being empty. For a
           multilingual field, being complete means that a value is present for
           every language.'''
        if not self.isMultilingual(obj):
            return Field.isCompleteValue(self, obj, value)
        # As soon as a given language value is empty, the global value is not
        # complete.
        if not value: return True
        for v in value.itervalues():
            if Field.isEmptyValue(self, obj, v): return
        return True

    def getDiffValue(self, obj, value, language):
        '''Returns a version of p_value that includes the cumulative diffs
           between successive versions. If the field is non-multilingual, it
           must be called with p_language being None. Else, p_language
           identifies the language-specific part we will work on.'''
        res = None
        lastEvent = None
        name = language and ('%s-%s' % (self.name, language)) or self.name
        for event in obj.workflow_history['appy']:
            if event['action'] != '_datachange_': continue
            if name not in event['changes']: continue
            if res is None:
                # We have found the first version of the field
                res = event['changes'][name][0] or ''
            else:
                # We need to produce the difference between current result and
                # this version.
                iMsg, dMsg = obj.getHistoryTexts(lastEvent)
                thisVersion = event['changes'][name][0] or ''
                comparator = HtmlDiff(res, thisVersion, iMsg, dMsg)
                res = comparator.get()
            lastEvent = event
        if not lastEvent:
            # There is no diff to show for this p_language.
            return value
        # Now we need to compare the result with the current version.
        iMsg, dMsg = obj.getHistoryTexts(lastEvent)
        comparator = HtmlDiff(res, value or '', iMsg, dMsg)
        return comparator.get()

    def getUnilingualFormattedValue(self, obj, value, layoutType='view',
          showChanges=False, userLanguage=None, language=None):
        '''If no p_language is specified, this method is called by
           m_getFormattedValue for getting a non-multilingual value (ie, in
           most cases). Else, this method returns a formatted value for the
           p_language-specific part of a multilingual value.'''
        if Field.isEmptyValue(self, obj, value) and not showChanges: return ''
        res = value
        if self.isSelect:
            if isinstance(self.validator, Selection):
                # Value(s) come from a dynamic vocabulary
                val = self.validator
                if self.isMultiValued():
                    return [val.getText(obj, v, self, language=userLanguage) \
                            for v in value]
                else:
                    return val.getText(obj, value, self, language=userLanguage)
            else:
                # Value(s) come from a fixed vocabulary whose texts are in
                # i18n files.
                _ = obj.translate
                if self.isMultiValued():
                    res = [_('%s_list_%s' % (self.labelId, v), \
                             language=userLanguage) for v in value]
                else:
                    res = _('%s_list_%s' % (self.labelId, value), \
                            language=userLanguage)
        elif (self.format == String.XHTML) and showChanges:
            # Compute the successive changes that occurred on p_value
            res = self.getDiffValue(obj, res, language)
        elif self.format == String.TEXT:
            if layoutType in ('view', 'cell'):
                res = Escape.xhtml(res, p=True)
        # If value starts with a carriage return, add a space; else, it will
        # be ignored.
        if isinstance(res, basestring) and \
           (res.startswith('\n') or res.startswith('\r\n')): res = ' ' + res
        return res

    def getFormattedValue(self, obj, value, layoutType='view',
                          showChanges=False, language=None):
        '''Be careful: p_language represents the UI language, while "languages"
           below represents the content language(s) of this field. p_language
           can be used, ie, to translate a Selection value.'''
        languages = self.getAttribute(obj, 'languages')
        if (len(languages) == 1) or isinstance(value, basestring):
            return self.getUnilingualFormattedValue(obj, value, layoutType,
                     showChanges, userLanguage=language)
            # Normally, p_value should not be a string if there is a single
            # language. This can happen in exceptional cases, ie, in a
            # object's history (data change), when an object was transmitted
            # from one App1 to App2, where a field is unilingual in App1 and
            # multilingual in App2.
        # Return the dict of values whose individual, language-specific values
        # have been formatted via m_getUnilingualFormattedValue.
        if not value and not showChanges: return value
        res = {}
        for lg in languages:
            if not value: val = ''
            else: val = value[lg]
            res[lg] = self.getUnilingualFormattedValue(obj, val, layoutType,
                                                       showChanges, language=lg)
        return res

    def getShownValue(self, obj, value, layoutType='view',
                      showChanges=False, language=None):
        '''Be careful: p_language represents the UI language, while "languages"
           below represents the content language(s) of this field. For a
           multilingual field, this method only shows one specific language
           part.'''
        languages = self.getAttribute(obj, 'languages')
        if len(languages) == 1:
            return self.getUnilingualFormattedValue(obj, value, layoutType,
                                             showChanges, userLanguage=language)
        if not value: return value
        # Try to propose the part that is in the user language, or the part of
        # the first content language else.
        lg = obj.getUserLanguage()
        if lg not in value: lg = languages[0]
        return self.getUnilingualFormattedValue(obj, value[lg], layoutType,
                                                showChanges, language=lg)

    def getSearchValue(self, form):
        '''Potentially apply a transform to search value in p_form'''
        r = Field.getSearchValue(self, form)
        if r and self.transform:
            r = self.applyTransform(r)
        return r

    def getSortValue(self, obj):
        '''Return the value of p_self on p_obj that must be used for sorting.
           While the raw p_value may be the value to use in most cases, it is
           not always true. For example, a string like "Gaëtan" could have
           "gaetan" as sort value.'''
        r = self.getValue(obj)
        if self.format == String.LINE:
            # Normalize the value
            r = sutils.normalizeText(r, dash=False, space=False)
        return r

    def getPlaceholder(self, obj):
        '''Returns a placeholder for the field if defined'''
        res = self.getAttribute(obj, 'placeholder') or ''
        if res == True:
            # A placeholder must be set, but we have no value. In this case, we
            # take the field label.
            res = obj.translate(self.labelId)
        return res

    def extractText(self, value, lower=True, dash=False):
        '''Extracts pure text from XHTML p_value'''
        extractor = XhtmlTextExtractor(lower=lower, dash=dash,
                                       raiseOnError=False)
        return extractor.parse('<p>%s</p>' % value)

    def getIndexValue(self, obj, forSearch=False):
        '''Pure text must be extracted from rich content; multilingual content
           must be concatenated.'''
        # Must we produce an index value?
        if not self.getAttribute(obj, 'mustIndex'):
            return ValidValue.forString(None, forSearch)
        isXhtml = self.format == String.XHTML
        if self.isMultilingual(obj):
            res = self.getValue(obj)
            if res:
                vals = []
                for v in res.itervalues():
                    if isinstance(v, unicode): v = v.encode('utf-8')
                    if isXhtml: vals.append(self.extractText(v))
                    else: vals.append(v)
                res = ' '.join(vals)
        else:
            res = Field.getIndexValue(self, obj, forSearch)
            if res and isXhtml: res = self.extractText(res)
        return ValidValue.forString(res, forSearch)

    def getPossibleValues(self, obj, withTranslations=False,
                          withBlankValue=False, blankLabel=None, className=None,
                          ignoreMasterValues=False):
        '''Returns the list of possible values for this field (only for fields
           with self.isSelect=True). If p_withTranslations is True, instead of
           returning a list of string values, the result is a list of tuples
           (s_value, s_translation). Moreover, p_withTranslations can hold a
           given language: in this case, this language is used instead of the
           user language. If p_withBlankValue is True, a blank value is
           prepended to the list, excepted if the type is multivalued. Used in
           combination with p_withTranslations being True, the i18n label for
           translating the blank value is given in p_blankLabel. If p_className
           is given, p_obj is the tool and, if we need an instance of
           p_className, we will need to use p_obj.executeQuery to find one.'''
        if not self.isSelect: raise Exception('This field is not a selection.')
        # Get the user language for translations, from p_withTranslations
        lg = isinstance(withTranslations, str) and withTranslations or None
        req = obj.REQUEST
        master = self.master
        aobj = obj.appy()
        if not ignoreMasterValues and master and callable(self.masterValue):
            # This field is an ajax-updatable slave. Get the master value...
            if master.valueIsInRequest(obj, req):
                # ... from the request if available
                requestValue = master.getRequestValue(obj)
                masterValues = master.getStorableValue(obj, requestValue,
                                                       complete=True)
            elif not className:
                # ... or from the database if we are editing an object
                masterValues = master.getValue(obj)
            else:
                # We don't have any master value
                masterValues = None
            # Get possible values by calling self.masterValue
            if masterValues:
                values = self.masterValue(aobj, masterValues)
            else:
                values = []
            # Manage parameter p_withTranslations
            if not withTranslations: res = values
            else:
                res = []
                for v in values:
                    res.append((v, self.getFormattedValue(obj,v,language=lg)))
        else:
            # Get the possible values from attribute "validator"
            if isinstance(self.validator, Selection):
                # We need to call self.methodName for getting the (dynamic)
                # values. If methodName begins with _appy_, it is a special Appy
                # method: we will call it on the Mixin (=p_obj) directly. Else,
                # it is a user method: we will call it on the wrapper
                # (p_obj.appy()). Some args can be hidden into p_methodName,
                # separated with stars, like in this example: method1*arg1*arg2.
                # Only string params are supported.
                methodName = self.validator.methodName
                # Unwrap parameters if any.
                if methodName.find('*') != -1:
                    elems = methodName.split('*')
                    methodName = elems[0]
                    args = elems[1:]
                else:
                    args = ()
                # On what object must we call the method that will produce the
                # values?
                if methodName.startswith('tool:'):
                    obj = obj.getTool()
                    methodName = methodName[5:]
                else:
                    # We must call on p_obj. But if we have something in
                    # p_className, p_obj is the tool and not an instance of
                    # p_className as required. So find such an instance.
                    if className:
                        brains = obj.executeQuery(className, maxResults=1,
                                                  brainsOnly=True)
                        if brains:
                            obj = brains[0]._unrestrictedGetObject()
                # Do we need to call the method on the object or on the wrapper?
                if methodName.startswith('_appy_'):
                    exec 'res = obj.%s(*args)' % methodName
                else:
                    exec 'res = obj.appy().%s(*args)' % methodName
                if not withTranslations: res = [v[0] for v in res]
                elif isinstance(res, list): res = res[:]
            else:
                # The list of (static) values is directly given in
                # self.validator.
                res = []
                for value in self.validator:
                    label = '%s_list_%s' % (self.labelId, value)
                    if withTranslations:
                        res.append( (value, obj.translate(label, language=lg)) )
                    else:
                        res.append(value)
        if (withBlankValue == 'forced') or \
           (withBlankValue and not self.isMultiValued()):
            # Create the blank value to insert at the beginning of the list
            if withTranslations:
                label = blankLabel or self.noValueLabel
                blankValue = ('', obj.translate(label, language=lg))
            else:
                blankValue = ''
            # Insert the blank value in the result
            if isinstance(res, tuple):
                res = (blankValue,) + res
            else:
                res.insert(0, blankValue)
        return res

    def validateValue(self, obj, value):
        if self.format == String.CAPTCHA:
            challenge = obj.REQUEST.SESSION.get('captcha', None)
            # Compute the challenge minus the char to remove
            i = challenge['number']-1
            text = challenge['text'][:i] + challenge['text'][i+1:]
            if value != text:
                return obj.translate('bad_captcha')
        elif self.isSelect:
            # Check that the value is among possible values
            possibleValues = self.getPossibleValues(obj,ignoreMasterValues=True)
            if isinstance(value, basestring):
                error = value not in possibleValues
            else:
                error = False
                for v in value:
                    if v not in possibleValues:
                        error = True
                        break
            if error: return obj.translate('bad_select_value')

    def applyTransform(self, value):
        '''Applies a transform as required by self.transform on single
           value p_value.'''
        if self.transform in ('uppercase', 'lowercase'):
            # For those transforms, I will remove any accent, because, most of
            # the time, if the user wants to apply such effect, it is for ease
            # of data manipulation, so I guess without accent.
            value = sutils.normalizeString(value, usage='noAccents')
        # Apply the transform
        method = String.transformMethods.get(self.transform)
        if method:
            exec 'value = value.%s()' % method
        return value

    def getTextareaStyle(self):
        '''When this widget must be rendered as an HTML field of type
           "textarea", get the content of its "style" attribute.'''
        r = 'text-transform:%s' % self.transform
        if isinstance(self.width, str):
            # The field width must be expressed here and we do not use
            # attribute "cols".
            r += ';width:%s' % self.width
        return r

    def getTextareaCols(self):
        '''When this widget must be rendered as an HTML field of type
           "textarea", get the content of its "cols" attribute.'''
        # Use this attribute only if width is expressed as an integer value
        return isinstance(self.width, int) and self.width or ''

    def getUnilingualStorableValue(self, obj, value):
        isString = isinstance(value, basestring)
        isEmpty = Field.isEmptyValue(self, obj, value)
        # Apply transform if required
        if isString and not isEmpty:
           value = self.applyTransform(value)
        # Clean XHTML strings
        if not isEmpty and (self.format == String.XHTML):
            # When image upload is allowed, ckeditor inserts some "style" attrs
            # (ie for image size when images are resized). So in this case we
            # can't remove style-related information.
            try:
                value = XhtmlCleaner().clean(value)
            except XhtmlCleaner.Error, e:
                # Errors while parsing p_value can't prevent the user from
                # storing it.
                pass
        # Clean TEXT strings
        if not isEmpty and (self.format == String.TEXT):
            value = value.replace('\r', '')
        # Truncate the result if longer than self.maxChars
        if isString and self.maxChars and (len(value) > self.maxChars):
            value = value[:self.maxChars]
        # Get a multivalued value if required
        if value and self.isMultiValued() and \
           (type(value) not in sutils.sequenceTypes):
            value = [value]
        return value

    def getStorableValue(self, obj, value, complete=False):
        # If no object is passed, assume the field is unilingual
        if obj is None:
            return self.getUnilingualStorableValue(obj, value)
        # Check unilinguality
        languages = self.getAttribute(obj, 'languages')
        if len(languages) == 1:
            return self.getUnilingualStorableValue(obj, value)
        # A multilingual value is stored as a dict whose keys are ISO 2-letters
        # language codes and whose values are strings storing content in the
        # language ~{s_language: s_content}~.
        if not value: return
        for lg in languages:
            value[lg] = self.getUnilingualStorableValue(obj, value[lg])
        return value

    def store(self, obj, value):
        '''Stores p_value on p_obj for this field'''
        languages = self.getAttribute(obj, 'languages')
        if (len(languages) > 1) and value and \
           (not isinstance(value, dict) or (len(value) != len(languages))):
            raise Exception('Multilingual field "%s" accepts a dict whose '\
                            'keys are in field.languages and whose ' \
                            'values are strings.' % self.name)
        Field.store(self, obj, value)

    def storeFromAjax(self, obj):
        '''Stores the new field value from an Ajax request, or do nothing if
           the action was canceled.'''
        rq = obj.REQUEST
        if rq.get('cancel') == 'True': return
        requestValue = rq['fieldContent']
        # Remember previous value if the field is historized
        isHistorized = self.getAttribute(obj, 'historized')
        previousData = None
        if isHistorized: previousData = obj.rememberPreviousData(self)
        if self.isMultilingual(obj):
            if isHistorized:
                # We take a copy of previousData because it is mutable (dict)
                prevData = previousData[self.name]
                if prevData != None: prevData = prevData.copy()
                previousData[self.name] = prevData
            # We get a partial value, for one language only
            language = rq['languageOnly']
            v = self.getUnilingualStorableValue(obj, requestValue)
            getattr(obj.aq_base, self.name)[language] = v
            part = ' (%s)' % language
        else:
            self.store(obj, self.getStorableValue(obj, requestValue))
            part = ''
        # Update the object history when relevant
        if isHistorized and previousData: obj.historizeData(previousData)
        # Update obj's last modification date
        from DateTime import DateTime
        obj.modified = DateTime()
        obj.reindex()
        obj.log('ajax-edited %s%s on %s.' % (self.name, part, obj.id))

    def getIndexType(self):
        '''Index type varies depending on String parameters.'''
        # If String.isSelect, be it multivalued or not, we define a ListIndex:
        # this way we can use AND/OR operator.
        if self.isSelect:
            return 'ListIndex'
        elif self.format == String.TEXT:
            return 'TextIndex'
        elif self.format == String.XHTML:
            return 'XhtmlIndex'
        return Field.getIndexType(self)

    def getJs(self, layoutType, res, config):
        if (self.format == String.XHTML) and (layoutType in ('edit', 'view')):
            # Compute the URL to ckeditor CDN
            ckUrl = String.cdnUrl % (config.ckVersion, config.ckDistribution)
            if ckUrl not in res: res.append(ckUrl)

    def getCaptchaChallenge(self, session, minLength=5, maxLength=9):
        '''Returns a Captcha challenge in the form of a dict. At key "text",
           value is a string that the user will be required to re-type, but
           without 1 character whose position is at key "number". The challenge
           is stored in the p_session, for the server-side subsequent check.'''
        # The challenge the user needs to type (minus one char)
        challenge = sutils.PasswordGenerator.get(minLength=minLength,
                                                 maxLength=maxLength)
        # Compute the position of the char to remove
        number = random.randint(1, len(challenge))
        r = {'text': challenge, 'number': number}
        session['captcha'] = r
        return r

    def generatePassword(self, maxLength=9):
        '''Generates a password (we recycle here the captcha challenge
           generator).'''
        return sutils.PasswordGenerator.get(maxLength=maxLength)

    ckLanguages = {'en': 'en_US', 'pt': 'pt_BR', 'da': 'da_DK', 'nl': 'nl_NL',
                   'fi': 'fi_FI', 'fr': 'fr_FR', 'de': 'de_DE', 'el': 'el_GR',
                   'it': 'it_IT', 'nb': 'nb_NO', 'pt': 'pt_PT', 'es': 'es_ES',
                   'sv': 'sv_SE'}
    def getCkLanguage(self, obj, language):
        '''Gets the language for CK editor SCAYT. p_language is one of
           self.languages if the field is multilingual, None else. If p_language
           is not supported by CK, we use english.'''
        if not language:
            language = self.getAttribute(obj, 'languages')[0]
        if language in self.ckLanguages: return self.ckLanguages[language]
        return 'en_US'

    def getCkParams(self, obj, language):
        '''Gets the base params to set on a rich text field'''
        tool = obj.getTool()
        base = tool.getSiteUrl()
        ckAttrs = {'customConfig': '%s/ui/ckeditor/config.js' % base,
                   'contentsCss': '%s/ui/ckeditor/contents.css' % base,
                   'stylesSet': '%s/ui/ckeditor/styles.js' % base,
                   'toolbar': self.toolbar, 'format_tags':';'.join(self.styles),
                   'scayt_sLang': self.getCkLanguage(obj, language)}
        if self.width: ckAttrs['width'] = self.width
        if self.height: ckAttrs['height'] = self.height
        if self.spellcheck: ckAttrs['scayt_autoStartup'] = True
        # Add custom styles
        if self.customStyles:
            for style in self.customStyles:
                id = style['id']
                ckAttrs['format_%s' % id] = style
                ckAttrs['format_tags'] += ';%s' % id
        if self.allowImageUpload:
            ckAttrs['filebrowserUploadUrl'] = '%s/upload' % obj.absolute_url()
        if not tool.getUser().has_role('Manager'):
            ckAttrs['removeButtons'] = 'Source'
        ck = []
        for k, v in ckAttrs.iteritems():
            if isinstance(v, int): sv = str(v)
            elif isinstance(v, bool): sv = str(v).lower()
            elif isinstance(v, dict): sv = str(v)
            else: sv = '"%s"' % v
            ck.append('%s: %s' % (k, sv))
        return ', '.join(ck)

    def getJsInit(self, obj, language):
        '''Gets the Javascript init code for displaying a rich editor for this
           field (rich field only). If the field is multilingual, we must init
           the rich text editor for a given p_language (among self.languages).
           Else, p_languages is None.'''
        name = not language and self.name or ('%s_%s' % (self.name, language))
        return 'CKEDITOR.replace("%s", {%s})' % \
               (name, self.getCkParams(obj, language))

    def getJsInlineInit(self, obj, name, language):
        '''Gets the Javascript init code for enabling inline edition of this
           field (rich text only). If the field is multilingual, the current
           p_language is given and p_name includes it. Else, p_language is
           None.'''
        id = obj.id
        fieldName = language and name.rsplit('_',1)[0] or name
        lg = language or ''
        return "CKEDITOR.disableAutoInline = true;\n" \
               "CKEDITOR.inline('%s_%s_ck', {%s, on: {blur: " \
               "function( event ) { var content = event.editor.getData(); " \
               "doInlineSave('%s','%s','%s','view',true,content,'%s')}}})" % \
               (id, name, self.getCkParams(obj, language), id, fieldName,
                obj.absolute_url(), lg)

    def isSelected(self, obj, fieldName, vocabValue, dbValue):
        '''When displaying a selection box (only for fields with a validator
           being a list), must the _vocabValue appear as selected? p_fieldName
           is given and used instead of field.name because it may be a a fake
           name containing a row number from a field within a list field.'''
        rq = obj.REQUEST
        # Get the value we must compare (from request or from database)
        if rq.has_key(fieldName):
            compValue = rq.get(fieldName)
        else:
            compValue = dbValue
        # Compare the value
        if type(compValue) in sutils.sequenceTypes:
            return vocabValue in compValue
        return vocabValue == compValue
# ------------------------------------------------------------------------------



'''An index for Text fields, splitting text into words'''

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
from appy.database.indexes import Index
from appy.utils.string import Normalize

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class TextIndex(Index):
    '''Index for Text fields'''

    # For a Text index, a value to store in the index is already and always
    # built as a tuple (by m_toIndexed hereabove): so it it always considered to
    # be "multiple", and it is useless to convert it to a tuple.
    def isMultiple(self, value, inIndex=False): return isinstance(value, tuple)
    def getMultiple(self, value): return value

    # A Text index potentially contains a large number of values. Comparing
    # values while reindexing an object may be inefficient, so here, we force
    # bypassing value comparison: everytime an object will be reindexed, all
    # existing values for a Text index will be removed and new ones will be
    # re-inserted.
    def valueEquals(self, value, current): return

    @classmethod
    def toIndexed(class_, value, field, normalize=True, ignore=2,
                  ignoreNumbers=False):
        '''Splits the plain text p_value into words'''
        # Words whose length is <= p_ignore are ignored, excepted, if
        # p_ignoreNumbers is False, words being numbers.
        if not value: return
        # Create a set
        r = set()
        if normalize:
            value = Normalize.text(value)
        for word in value.split():
            # Keep this word or not ?
            if len(word) <= ignore:
                keepIt = not ignoreNumbers and word.isdigit()
            else:
                keepIt = True
            if keepIt:
                r.add(word)
        return tuple(r)

    # No need to override m_toString in order to normalize the value: it will be
    # done by the global "searchable" index.

    @classmethod
    def toTerm(class_, value, field):
        '''Normalizes p_value in order to be used as a search term'''
        return Normalize.text(value)
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



'''Meta-class for a Appy class'''

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
import collections
import persistent
from persistent.mapping import PersistentMapping


from appy.px import Px
from appy.model import Model

import appy.model.meta.class_
from appy.model.meta import Meta
from appy.model.base import Base
from appy.model.page import Page
from appy.model.user import User
from appy.model.tool import Tool
from appy.model.group import Group
from appy.ui.layout import Layouts
from appy.model.fields import Field
from appy.model.workflow import Role
from appy.model.searches import Search
from appy.utils import string as sutils
from appy.model.document import Document
from appy.model.utils import Object as O
from appy.model.fields.phase import Phase
from appy.model.fields.string import String
from appy.model.translation import Translation
from appy.model.workflow.history import History
from appy.model.searches.gridder import Gridder
from appy.model.fields.computed import Computed
from appy.model.workflow.localRoles import LocalRoles
from appy.model.fields.phase import Page as FieldPage

# Errors - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
UNALLOWED_FIELD_NAME = 'You have defined a field or search named "%s" on ' \
  'class "%s". This name is not allowed.'
CREATOR_LOCAL_ROLE = '%s: local role "%s" cannot be used as creator.'
DUPLICATE_SEARCH_NAME = '%s: duplicate search "%s".'
SEARCH_FIELD_HOMONYM = '%s: "%s" is both used as search and field name, this ' \
  'is not allowed.'
CUSTOM_ID_NOT_STRING = 'The custom ID produced by method "generaeId" must be ' \
  'a string.'

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Class(Meta):
    '''Represents an Appy class'''

    # Attributes added on Appy classes
    attributes = {'fields': Field}

    # Reserved attribute names unallowed for naming a Field or a Search
    unallowedNames = {
      'id':         None, # The object's identifier
      'iid':        None, # The object's integer identifier
      'values':     None, # Dict storing all field values
      'container':  None, # Info about the objet's container
      'history':    None, # The object's history
      'localRoles': None, # The object's local roles
      'locks':      None, # Locks on object pages
    }

    # Fields being unwanted on search fields, although being indexed. Field
    # "title" is not really "unsearchable", but in most cases, it is preferable
    # to use field "searchable".
    unsearchableFields = ('title', 'allowed', 'cid')

    @classmethod
    def unallowedName(class_, name):
        '''Return True if p_name can't be used for naming a Field or Search'''
        # A name is unallowed if
        # (a) it corresponds to some standard attribute added at object creation
        #     on any Appy object (but not declared at the class level);
        # (b) it corresponds to an attribute or method defined in class Base.
        return (name in class_.unallowedNames) or (name in Base.__dict__)

    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # Model-construction-time methods
    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # The following set of methods allow to build the model when making an app
    # or when starting the server.
    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    def __init__(self, class_, isBase=False, appOnly=False):
        Meta.__init__(self, class_, appOnly)
        # p_class_ may be of 3 types:
        # ----------------------------------------------------------------------
        # "base" | A class from the base Appy model (from appy.model) that has
        #        | not been overridden by an app's class.
        # ----------------------------------------------------------------------
        # "app"  | An original class from the app, that does not exist in the
        #        | base model.
        # ----------------------------------------------------------------------
        # "over" | A class defined in the app but that overrides a class defined
        #        | in the base Appy model.
        # ----------------------------------------------------------------------
        self.type = 'base' if isBase else 'app'
        # Create the concrete class from which instances will be created.
        # Indeed, In order to bring all Appy functionalities to p_class_, we
        # need to create a new class inheriting from:
        # ----------------------------------------------------------------------
        #   p_class_  | The class defined by the developer or being part of the
        #             | base Appy model.
        # ----------------------------------------------------------------------
        #    Base*    | Class appy.model.base.Base, the base class for any Appy
        #             | class, or one if its sub-classes, depending on p_class_
        #             | name. For example, if the user defines a class named
        #             | "Invoice", it will inherit from Base (=a class of type
        #             | "app"). If the user defines a class named "User", it
        #             | will inherit from class appy.model.user.User (= a class
        #             | of type "over"): using the name of a base class from
        #             | package appy.model implicitly means that the developer
        #             | wants to extend this class.
        #             |
        #             | Note that for not-overridden classes from appy.model (=
        #             | classes of type "base"), there is no base class: the
        #             | class is already a base class.
        # ----------------------------------------------------------------------
        #  Persistent | Class persistent.Persistent allows to store instances in
        #             | the database (ZODB).
        # ----------------------------------------------------------------------
        # Creating the concrete class is not necessary when p_self.appOnly is
        # True.
        if not appOnly:
            # The name of the generated class (tribute to appy.gen)
            genName = 'Gen_%s' % self.name
            if self.type == 'base':
                # The class must only inherit from p_class_ and Persistent
                exec('class %s(class_, persistent.Persistent): pass' % genName)
            else:
                # It must also inherit from a base class or one of its
                # sub-classes.
                base = self.getBaseClass()
                if base != 'Base':
                    # This is an "over" class
                    self.type = 'over'
                exec('class %s(class_, %s, persistent.Persistent): pass' % \
                     (genName, base))
            # This concrete class will have a full name being
            #                appy.model.meta.class_.<genName>
            exec('self.concrete = %s' % genName)
            # Add the class in the namespace of the current package
            # (appy.model.meta.class_). Else, pickle, used by the ZODB to store
            # instances of this class, will not find it.
            setattr(appy.model.meta.class_, genName, self.concrete)
        else:
            self.type = (self.name in Model.baseClasses) and 'over' or 'app'
        # Fields are grouped into pages, themselves grouped into "phases"
        self.phases = collections.OrderedDict()
        # Read fields and static searches
        self.readFields()
        self.readSearches()
        # The class will be linked to a workflow (later, by the loader)
        self.workflow = None
        # Call the "update" method if present: it allows developers to modify
        # this class. One of the most frequent updates will be to update field
        # "title".
        if self.type == 'over':
            # Execute the "update" method on the base class
            base = eval(self.name)
            if hasattr(base, 'update'):
                base.update(self)
        if hasattr(self.python, 'update'): self.python.update(self)

    def getBaseClass(self):
        '''Returns the name of the class being p_self's base class'''
        name = self.name
        return name if name in Model.baseClasses else 'Base'

    def injectProperties(self):
        '''Injects, on p_self.concrete, a property for every field defined on
           this class, in order to get and set its value.'''
        # Other techniques could have been implemented for injecting such
        # virtual attributes, like overriding __getattribute__ and __setattr__,
        # but this should be avoided in conjunction with ZODB and pickle.
        class_ = self.concrete
        # Inject one property for every field
        for name, field in self.fields.items():
            # To get a field value is done via method Field::getValue
            getter = lambda self, field=field: field.getValue(self)    
            # To set a value is done via method Field::store
            if field.type == 'Ref':
                setter = lambda self, v, field=field: \
                                field.store(self, v, secure=False)
            else:
                setter = lambda self, v, field=field: field.store(self, v)
            setattr(class_, name, property(getter, setter))

    def getFieldClasses(self, class_):
        '''Returns, on p_class_ and its base classes, the list of classes where
           fields may be defined.'''
        # Add p_class_ in itself
        r = [class_]
        # Scan now p_class_'s base class
        if (class_ != self.python) or (self.type == 'base'):
            base = class_.__bases__[0]
        else:
            base = eval(self.getBaseClass())
        if base.__name__ != 'object':
            r  = self.getFieldClasses(base) + r
        return r

    def readPage(self, page):
        '''While browsing a field, a p_page has been encountered. Add it to
           self.phases if not already done.'''
        # Create the phase when appropriate
        name = page.phase
        if name not in self.phases:
            phase = self.phases[name] = Phase(name)
        else:
            phase = self.phases[name]
        # Add the page into it
        if page.name not in phase.pages:
            phase.pages[page.name] = page

    def readFields(self):
        '''Create attribute "fields" as an ordered dict storing all fields
           defined on this class and parent classes.'''
        class_ = self.python
        r = collections.OrderedDict()
        Err = Model.Error
        # Find field in p_class_ and base classes
        for aClass in self.getFieldClasses(class_):
            for name, field in aClass.__dict__.items():
                if name.startswith('__'): continue
                if isinstance(field, Field):
                    # Ensure the field name is acceptable
                    self.checkAttributeName(name)
                    if not aClass.__module__.startswith('appy.model') and \
                       (Class.unallowedName(name)):
                        raise Err(UNALLOWED_FIELD_NAME % (name, self.name))
                    # A field was found.
                    # ~~~
                    # Special fields "title" and "record" must be duplicated.
                    # Else, all app classes will get the modifications that the
                    # developer can apply to it via "update" methods.
                    if name == 'title':
                        field = String(**Base.titleAttributes)
                    elif name == 'record':
                        field = Computed(**Base.recordAttributes)
                        field.page = field.page.clone()
                    # Late-initialize it
                    field.init(self, name)
                    r[name] = field
                    if field.page:
                        self.readPage(field.page)
        self.fields = r

    def readSearches(self):
        '''Create attribute "searches" as an ordered dict storing all static
           searches defined on this class.'''
        class_ = self.python
        r = collections.OrderedDict()
        Err = Model.Error
        if hasattr(class_, 'searches'):
            for search in class_.searches:
                name = search.name
                # Ensure the search name is acceptable
                if Class.unallowedName(name):
                    raise Err(UNALLOWED_FIELD_NAME % (name, self.name))
                # Ensure the search name is unique, also among fields
                if name in r:
                    raise Err(DUPLICATE_SEARCH_NAME % (self.name, name))
                if name in self.fields:
                    raise Err(SEARCH_FIELD_HOMONYM % (self.name, name))
                # Ensure the field name is acceptable
                self.checkAttributeName(name)
                search.init(self)
                r[search.name] = search
        self.searches = r

    def getCreators(self):
        '''Gets the roles allowed to create instances of this class'''
        r = []
        class_ = self.python
        # The info can be defined in a static attribute named "creators"
        creators = getattr(self.python, 'creators', None)
        if not creators: return r
        # Athough a method can be specified in "creators", we can't execute it,
        # so we care about creators only if defined "statically", as a list.
        if not isinstance(creators, list): return r
        # Browse roles
        for creator in creators:
            if isinstance(creator, Role):
                if creator.local:
                    error = CREATOR_LOCAL_ROLE % (self.name,creator.name)
                    raise Model.Error(error)
                r.append(creator)
            else:
                r.append(Role(creator))
        return r

    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # Instance-creation method
    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    def new(self, iid, id, login, initialComment=None, initialState=None):
        '''Create and return an instance of the concrete class'''
        o = self.concrete()
        # Set the IDs on the object
        o.iid = iid
        o.id = id
        # Create the object's history
        o.history = History(o)
        # Create a dict storing the object's local roles, and grant role 'Owner'
        # to the user behind this object creation, whose p_login is given.
        o.localRoles = LocalRoles()
        o.localRoles.add(login, 'Owner')
        # Create the dict "values" that will store field values, and set the
        # first values in it.
        o.values = PersistentMapping() # ~{s_fieldName: fieldValue}~
        # Initialise object's history by triggering the _init_ transition
        tr = self.workflow.initialTransition
        state = initialState or tr.states[1]
        tr.trigger(o, initialComment, doSay=False, forceTarget=state)
        return o

    def generateId(self, o):
        '''A method named "generateId" may exist on p_self.python, allowing the
           app's developer to choose itself an ID for the p_o(bject) under
           creation. This method accepts p_o as unique arg and must return the
           ID as a string, or None.'''
        # Return None if no such method is defined on p_self.python
        if not hasattr(self.python, 'generateId'): return
        r = self.python.generateId(o)
        # If the method return None, a standard ID will be generated by Appy
        if r is None: return
        # Raise an error if the returned ID is not a string
        if not isinstance(r, str):
            raise Exception(CUSTOM_ID_NOT_STRING)
        return r

    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # Run-time methods
    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # The following set of methods allow to compute, at run-time,
    # class-dependent elements.
    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    def getListCss(self, layout='view'):
        '''Returns the CSS class(es) to apply when displaying a list (=a table)
           containing instances of this class.'''
        css = getattr(self.python, 'listCss', None)
        return css if css else 'list'

    def getListTitle(self, o):
        '''Gets p_o's title as it must appear within lists of objects'''
        if hasattr(self.python, 'listTitle'):
            r = self.python.listTitle(o, nav)
        else:
            r = o.getShownValue('title')
        return r

    def getPageLayouts(self, o):
        '''Gets the page layouts being applicable to p_o, being an instance of
           p_self.'''
        r = getattr(self.python, 'layouts', None)
        if callable(r): r = r(o)
        return r or Layouts.Page.defaults

    def produceUniqueLinks(self):
        '''When producing links ("a" tags) to instances of this class, must
           these links always be unique ?'''
        return hasattr(self.python, 'uniqueLinks') and self.python.uniqueLinks

    def showTransitions(self, o, layout):
        '''When displaying p_self's instance p_o, must we show, on p_layout, the
           buttons/icons for triggering transitions ?'''
        # Never show transitions on edit pages
        if layout == 'edit': return
        # Use the default value if self does not specify it
        if not hasattr(self.python, 'showTransitions'): return layout == 'view'
        r = self.python.showTransitions
        r = r(o) if callable(r) else r
        # This value can be a single value or a tuple/list of values
        return layout == r if isinstance(r, str) else layout in r

    def getCssFor(self, o, elem):
        '''Gets the name of the CSS class to use for styling some p_elem on
           instance p_o. If this class does not define a dict or method named
           "styles", the defaut CSS class to use will be named p_elem.'''
        styles = getattr(self.python, 'styles', None)
        if not styles: return elem
        elif callable(styles): # A method
            return styles(o, elem) or elem
        else: # A dict or a appy.Object instance
            return styles.get(elem) or elem

    def getCreateVia(self, tool):
        '''How to create instances of this class ? The answer can be: via a
           web form ("form"), via a form pre-filled from a template (a Search
           instance) or programmatically only (None).'''
        # By default, instances are created via an empty web form
        r = getattr(self.python, 'createVia', 'form')
        return r if not callable(r) else r(tool)

    def getCreateLink(self, tool, create, formName,
                      sourceField=None, insert=None):
        '''When instances of p_self must be created from a template (from the
           UI), this method returns the link allowing to query such templates
           from a popup.'''
        r = '%s/query?class=%s&search=fromSearch&popup=1&fromClass=%s' \
            '&formName=%s' % (tool.url, create.class_.__name__, self.name,
                              formName)
        # When object creation occurs via a Ref field, its coordinates are given
        # in p_sourceField as a string: "sourceObjectId:fieldName".
        if sourceField: r += '&sourceField=%s' % sourceField
        # p_insert is filled when the object must be inserted at a given place
        if insert: r += '&insert=%s' % insert
        return r

    def getCreateExclude(self, o):
        '''When an instance of this class must be created from a template
           object, what fields must not be copied from the template ?'''
        r = getattr(self.python, 'createExclude', ())
        return r(o) if callable(r) else r

    def getConfirmPopupWidth(self, popup):
        '''Gets the width (in pixels, as an integer value), of the confirmation
           popup possibly shown when creating or editing p_self's instances.'''
        return getattr(self.python, 'confirmPopup', 350 if popup else 500)

    def maySearch(self, tool, layout):
        '''May the user search among instances of this class ?'''
        # When editing a form, one should avoid annoying the user with this
        if layout == 'edit': return
        return self.python.maySearch(tool) \
               if hasattr(self.python, 'maySearch') else True

    def getSearchAdvanced(self, tool):
        '''Gets the "advanced" search defined on this class if present'''
        # This Search instance, when present, is used to define search
        # parameters for this class' special searches: advanced, all and live
        # searches.
        r = getattr(self.python, 'searchAdvanced', None)
        return r if not callable(r) else r(tool)

    def maySearchAdvanced(self, tool):
        '''Is advanced search enabled for this class ?'''
        # Get the "advanced" search
        advanced = self.getSearchAdvanced(tool)
        # By default, advanced search is enabled
        if not advanced: return True
        # Evaluate attribute "show" on this Search instance representing the
        # advanced search.
        return advanced.isShowable(tool)

    def isIndexable(self):
        '''Is p_self "indexable" ? In other words: do we need to have a Catalog
           storing index values for instances of this class ?'''
        return getattr(self.concrete, 'indexable', True)

    def getListPods(self):
        '''Finds, among p_self' fields, those being Pod fields showable on
           search results.'''
        return [f for f in self.fields.values() \
                if (f.type == 'Pod') and (f.show == 'query')]

    def getDynamicSearches(self, tool):
        '''Gets the dynamic searches potentially defined on this class'''
        # Stop here if no dynamic search is defined
        if not hasattr(self.python, 'getDynamicSearches'): return
        # Create the cache if it does not exist yet
        handler = tool.H()
        dyn = handler.cache.dynamicSearches
        if dyn is None:
            dyn = handler.cache.dynamicSearches = {}
        # Get the cached searches when present; compute and cache them else
        name = self.name
        if name in dyn:
            r = dyn[name]
        else:
            r = dyn[name] = self.python.getDynamicSearches(tool)
            for search in r: search.init(self)
        return r

    def getSearch(self, name, tool=None):
        '''Returns static or dynamic search whose name is p_name'''
        # Search among static searches
        r = self.searches.get(name)
        if r or (tool is None): return r
        # Search among dynamic searches
        dyn = self.getDynamicSearches(tool)
        if dyn:
            for search in dyn:
                if search.name == name:
                    return search

    def getGroupedSearches(self, tool, ctx):
        '''Returns an object with 2 attributes:
           * "searches" stores the searches defined for this class, as instances
             of the run-time-specific class appy.model.searches.UiSearch;
           * "default" stores the search being the default one.
        '''
        searches = []
        default = None # Also retrieve the default one here
        groups = {} # The already encountered groups
        page = FieldPage('searches') # A dummy page required by class UiGroup
        # Get the statically defined searches from class's "searches" attribute
        for search in self.searches.values():
            # Ignore search that can't be shown
            if not search.isShowable(tool): continue
            searches.append(search)
        # Get the dynamically computed searches
        dyn = self.getDynamicSearches(tool)
        if dyn: searches += dyn
        # Return the grouped list of UiSearch instances
        r = []
        for search in searches:
            # Create the search descriptor
            ui = search.ui(tool, ctx)
            if not search.group:
                # Insert the search at the highest level, not in any group
                r.append(ui)
            else:
                uiGroup = search.group.insertInto(r, groups, page, self.name,
                                                  content='searches')
                uiGroup.addElement(ui)
            # Is this search the default search?
            if search.default: default = ui
        return O(all=r, default=default)

    def getSearchFields(self, tool, refInfo=None):
        '''Returns, as 2-tuple:
           - p_self's searchable fields (some among all indexed fields),
           - CSS files required to render the search fields,
           - Javascript files required to render these fields,
           - a Gridder instance determining how to render the search form.'''
        # We will collect CSS and JS files required for search fields
        fields = []; css = []; js = []
        if refInfo:
            # The search is triggered from a Ref field
            o, ref = Search.getRefInfo(tool, refInfo, nameOnly=False)
            names = ref.queryFields or ()
            cols = ref.queryNbCols
        else:
            # The search is triggered from an app-wide search
            names = getattr(self.python, 'searchFields', None)
            names = names(tool) if callable(names) else names
            cols = getattr(self.python, 'numberOfSearchColumns', 1)
        # Get fields from names
        if names:
            fields = [self.fields.get(name) for name in names]
        else:
            # If fields are not explicitly listed, take all indexed fields,
            # "title" excepted, "searchable" and "title" being mostly redundant.
            fields = [field for field in self.fields.values() \
             if field.indexed and (field.name not in Class.unsearchableFields)]
        # Collect CSS and JS files
        config = tool.config
        for field in fields:
            field.getCss('edit', css)
            field.getJs('edit', js, config)
        # Returns a gridder allowing to render the search field
        gridder = Gridder(cols=cols)
        return fields, css, js, gridder

    def getResultModes(self):
        '''Gets the search result modes for this class'''
        return getattr(self.python, 'resultModes', None)

    def getResultsTop(self, tool, search, mode):
        '''Returns something to display on the results page just before
           displaying search results.'''
        method = getattr(self.python, 'getResultsTop', None)
        return method(tool, search, mode) if method else None

    def getResultCss(self, layout='view'):
        '''Gets the CSS class(es) to apply when displaying a table containing
           p_self's instances.'''
        r = getattr(self.python, 'listCss', None)
        return r if r else 'list'

    def getTitleMode(self):
        '''Gets the "title mode" for instances of this class'''
        # Consult homonym method on class appy.model.searches.UiSearch
        return getattr(self.python, 'titleMode', None)

    def hasSubTitle(self):
        '''Does p_self define a method "getSubTitle"?'''
        # This method allows to fill a zone under the title of p_self's
        # instances when rendered in a table.
        return hasattr(self.python, 'getSubTitle')

    def getPortletBottom(self, tool):
        '''Is there a custom zone to display at the bottom of the portlet zone
           for this class ?'''
        if not hasattr(self.python, 'getPortletBottom'): return
        return self.python.getPortletBottom(tool)

    def getListColumns(self, tool):
        '''Return the list of columns to display when rendering p_self's
           instances in a table.'''
        r = getattr(self.python, 'listColumns', ('title',))
        return r if not callable(r) else r(tool)

    def getFilters(self, tool, filters=None):
        '''Extracts, from the request (or from p_filters if passed), filters
           defined on fields belonging to this class.'''
        r = sutils.getDictFrom(filters or tool.req.filters)
        # Apply a potential transform on every filter value
        for name in list(r.keys()):
            # Get the corresponding field
            field = self.fields.get(name)
            if field:
                try:
                    value = field.getStorableValue(None, r[name])
                    r[name] = field.getFilterValue(value)
                except Exception:
                    # The encoded value is invalid. Ignore it.
                    del(r[name])
        return r

    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # PXs
    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    # Icon for creating instances of this class from a template
    pxAddFrom = Px('''
     <a target="appyIFrame" id=":addFormName + '_from'"
        href=":class_.getCreateLink(tool, create, addFormName, sourceField)">
      <input var="css='Small' if fromRef else 'Portlet';
                  label=_('object_add_from')"
         type="button" value=":label" class=":'button%s button' % css"
         onclick="openPopup('iframePopup')" style=":url('add.svg', bg=True)"/>
     </a>''')

    # Style characteristics to apply to the "add" button, depending on its type
    addStyles = {
      # The "add" button as it appears in the portlet
      'portlet': O(css='buttonPortlet button', bgIcon='portletAdd.svg',
                   bgSize='24px 24px'),
      # The "add" button as it may appear elsewhere (ie, on search results)
      'small'  : O(css='buttonSmall button', bgIcon='add.svg',
                   bgSize='18px 18px'),
      # When using pxAdd from a custom PX, let this latter manage button style
      None     : O(css='', bgIcon=None, bgSize=None)
    }

    # Form for creating instances of this class from the portlet or a search
    pxAdd = Px('''
     <form var="create=class_.getCreateVia(tool); className=class_.name"
           if="create" class="addForm" name=":'%s_add' % className"
           var2="styles=class_.addStyles[buttonType];
                 target=ui.LinkTarget(class_.python, popup=viaPopup);
                 text=_(label or 'object_add')"
           action=":'%s/new' % tool.url" target=":target.target">
      <input type="hidden" name="className" value=":className"/>
      <input type="hidden" name="template" value=""/>
      <input type="hidden" name="insert" value=""/>
      <input type="hidden" name="nav" value=":nav"/>
      <input type="hidden" name="popup"
           value=":'True' if (popup or (target.target!='_self')) else 'False'"/>

      <!-- Create from an empty form -->
      <input class=":styles.css" value=":text" title=":text" type="submit"
        style=":url(styles.bgIcon, bg=styles.bgSize) if styles.bgIcon else ''"
        onclick=":target.getOnClick('searchResults')"/>

      <!-- Create from a pre-filled form when relevant -->
      <div if="create != 'form'" class="addFrom"
         var2="fromRef=False; sourceField=None;
               addFormName='%s_add' % className">:class_.pxAddFrom</div>
     </form>''',

     css='''.addFrom { margin-top: 5px }''')
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



'''Manages links between objects while importing a web of objects from one site
   to another.'''

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

# During the process of importing of a web of interconnected objects from one
# site to another, at a given moment in time, an object may need to be linked to
# another one that has not been imported yet.

# Class "Unresolved" stores those "unresolved" links, that are collected in a
# first import step. In a second step, once the missing objects will have been
# imported, unresolved links will actually be "reified".

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
from appy.model.utils import Object as O

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
RES_START   = 'Resolving links for %d objects, based on a total of %d ' \
              'imported objects...'
RES_END     = '%d links resolved and %d unresolved.'
UNRESOLVED  = 'Unresolved object with ID "%s" mentioned in %s:%s:%s.'
MIS_CUS_FUN = 'Custom entries to resolve have been found but no function has ' \
              'been specified to resolve them (in peer.unresolved.customFun).'
CUS_START   = 'Starting custom resolution of %d entries...'
CUS_END     = 'Custom resolution done.'
FIN_START   = 'Resolver: calling the "final" function...'
FIN_END     = 'Final job done.'

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Unresolved:
    '''Manage unresolved links between objects'''

    def __init__(self, peer):
        self.peer = peer
        # Dict "unresolved" stores, for every local object and every forward Ref
        # defined on it, the list of distant IDs for which no local object
        # exists yet.
        self.unresolved = {} # ~{Base: {s_FwdRefFieldName: [s_distantId]}}~
        # Beyond this dict of "standard" unresolved entries, an app may store
        # custom data structures in Custom fields, requiring specific
        # resolution. For such cases, user the following dict.
        # - It must be filled by a custom Importer (appy.peer.importer.Importer
        #   sub-class) defined in your app and registered in the peer's
        #   "importers" attribute.
        # - Everytime your custom importer finds the content of a custom field
        #   containing elements to be resolved at the end of the process, it
        #   must add or update an entry in dict "custom" below, via method
        #   m_addCustom below.
        self.custom = {} # ~{Base: [s_customFieldName]}~
        # - At the end of the import process, after having performed standard
        #   resolving using p_self.unrevolved, custom resolving will be
        #   launched. For every object and custom field stored in p_self.custom,
        #   the custom function you must specify in the following attribute,
        #   will be called, with, as args:
        #  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        #  peer  | The currently runnning Peer instance ;
        #  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        #  o     | one of the objects from p_self.custom ;
        #  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        #  name  | the name of one of the custom fields mentioned in
        #        | p_self.custom at key "o".
        #  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        self.customFun = None
        # At the very end of the resolving process, a custom method can be ran.
        # It will receive the peer instance and the tool as args.
        self.finalFun = None

    def add(self, o, field, distantId):
        '''Adds an unresolved link between this local p_o(object) and a missing
           distant object whose ID is p_distantId, via this forward Ref
           p_field.'''
        # Get or create the sub-dict related to p_o
        if o in self.unresolved:
            sub = self.unresolved[o]
        else:
            sub = self.unresolved[o] = {}
        # Insert the ID in the appropriate list
        name = field.name
        if name in sub:
            sub[name].append(distantId)
        else:
            sub[name] = [distantId]

    def addCustom(self, o, field):
        '''Adds a custom entry in self.custom'''
        name = field.name
        if o in self.custom:
            self.custom[o].append(name)
        else:
            self.custom[o] = [name]

    def resolve(self, tool):
        '''Reify links between objects for which links could not be done earlier
           in the import process.'''
        counts = O(resolved=0, unresolved=0)
        peer = self.peer
        byId = peer.localObjects
        tool.log(RES_START % (len(self.unresolved), len(byId)))
        # Browse links to resolve, object per object
        for o, links in self.unresolved.items():
            # Browse links to resolve, Ref by Ref starting from this object
            for name, distantIds in links.items():
                for id in distantIds:
                    # Find the local object corresponding to this ID
                    tied = byId.get(id)
                    if tied is None:
                        # Still unresolved
                        counts.unresolved += 1
                        tool.log(UNRESOLVED % (id, o.class_.name, o.id, name))
                    else:
                        # Tied object found, perform the link
                        o.link(name, tied, executeMethods=False)
                        counts.resolved += 1
        tool.log(RES_END % (counts.resolved, counts.unresolved))
        # Perform custom resolution when relevant
        if self.custom:
            if not self.customFun:
                raise Exception(MIS_CUS_FUN)
            tool.log(CUS_START % len(self.custom))
            for o, names in self.custom.items():
                for name in names:
                    self.customFun(peer, o, name)
            tool.log(CUS_END)
        # Call the final function if present
        if self.finalFun:
            tool.log(FIN_START)
            self.finalFun(peer, tool)
            tool.log(FIN_END)
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
